# %% [markdown]
# ### Example script for reading and lightly processing DGS laptop data from a Thompson cruise, and comparing to BGM serial files.
#
# Data files are downloaded by the script using pooch
#
# Read DGS and navigation files
#
# Read BGM files
#
# Correct for meter bias with info from shipgrav
#
# Compare BGM and DGS outputs
#
# Use timestamps to sync more accurate nav with the gravity data.
#
# Calculate FAA (free air anomaly)
#
# Plot
#
# satellite data tracked from v32.1 Global Gravity grid, which
# includes data from SIO, NOAA, and NGA.
#
# Reference: Sandwell et al. (2014) New global marine gravity model
# from CryoSat-2 and Jason-1 reveals buried tectonic struture.
# Science 346(6205), DOI: 10.1126/science.1258213

# %%
import os
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pooch
import shipgrav.grav as sgg
import shipgrav.io as sgi
import tomli as tm
from scipy.interpolate import interp1d
from scipy.signal import filtfilt, firwin


# %%
# set some general metadata
ship = 'Thompson'
cruise = 'TN400'        # this is used for filepaths
sampling = 1            # 1 Hz - data should be at this rate already

# read a few constants etc from our toml database file
with open('../shipgrav/database.toml', 'rb') as f:
    info = tm.load(f)
nav_tag = info['nav-talkers'][ship]
biases = info['bias-values'][ship]

# %%
# get files: DGS laptop from zenodo, BGM serial and nav from R2R
dgs_files = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Thompson/TN400/gravimeter/DGS/AT1M-Grav-PROC_20220314-000001.Raw',
                    'data/Thompson/TN400/gravimeter/DGS/AT1M-Grav-PROC_20220313-000001.Raw']))

bgm_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/TN400/fileset/151470",
        known_hash='b0ccc52e60334284dd79573e27e050eead77f5ac1d59f510385c7e05a8474bcf',progressbar=True,
        processor=pooch.Untar(
            members=['TN400/151470/data/BGM3-GRAV-RAW_20220313-000001.Raw',
                    'TN400/151470/data/BGM3-GRAV-RAW_20220314-000001.Raw']))

nav_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/TN400/fileset/151457",
        known_hash="76e66365c41d393510bb7ab9637305296282e9041415c1343faa171af28abf85",progressbar=True,
        processor=pooch.Untar(
            members=['TN400/151457/data/POSMV-V5-INGGA-RAW_20220313-000001.Raw',
                    'TN400/151457/data/POSMV-V5-INGGA-RAW_20220314-000001.Raw']))

# %%
# read and sort the nav data
gps_nav = sgi.read_nav(ship, nav_files)
gps_nav.sort_values('time_sec', inplace=True)
gps_nav.reset_index(inplace=True, drop=True)

# %%
# we happen to know that there are some weird nav dropouts in this dataset
# so clean them up here
bad_inds = np.where(np.diff(gps_nav['lon']) > 1)[0]
gps_nav.drop(bad_inds, axis=0, inplace=True)

# %%
# read and sort the DGS laptop data
dgs_data = sgi.read_dgs_laptop(dgs_files, ship)
dgs_data.sort_values('date_time', inplace=True)
dgs_data.reset_index(inplace=True, drop=True)
dgs_data['tsec'] = [e.timestamp()
                    for e in dgs_data['date_time']]  # get posix timestamps
dgs_data['grav'] = dgs_data['rgrav'] + biases['dgs']

# %%
# read and sort the BGM data
bgm_data = sgi.read_bgm_raw(bgm_files, ship)
bgm_data.sort_values('date_time', inplace=True)
bgm_data.reset_index(inplace=True, drop=True)
bgm_data['tsec'] = [e.timestamp() for e in bgm_data['date_time']]
bgm_data['grav'] = bgm_data['rgrav'] + biases['bgm']

# %%
# sync data geographic coordinates to nav by interpolating with timestamps
# (interpolators use posix timestamps, not datetimes)
gps_lon_int = interp1d(gps_nav['time_sec'].values, gps_nav['lon'].values,
                       kind='linear', fill_value='extrapolate')
gps_lat_int = interp1d(gps_nav['time_sec'].values, gps_nav['lat'].values,
                       kind='linear', fill_value='extrapolate')
dgs_data['lon_new'] = gps_lon_int(dgs_data['tsec'].values)
dgs_data['lat_new'] = gps_lat_int(dgs_data['tsec'].values)

bgm_data['lon_new'] = gps_lon_int(bgm_data['tsec'].values)
bgm_data['lat_new'] = gps_lat_int(bgm_data['tsec'].values)

# %%
# calculate corrections for FAA for DGS and BGM
for df in [dgs_data, bgm_data]:
    ellipsoid_ht = np.zeros(len(df))  # we are working at sea level
    lat_corr = sgg.wgs_grav(df['lat_new']) + \
        sgg.free_air_second_order(df['lat_new'], ellipsoid_ht)
    eotvos_corr = sgg.eotvos_full(df['lon_new'].values, df['lat_new'].values,
                                  ellipsoid_ht, sampling)
    tide_corr = sgg.longman_tide_prediction(
        df['lon_new'], df['lat_new'], df['date_time'])

    df['faa'] = df['grav'] - lat_corr + eotvos_corr + tide_corr
    df['full_field'] = df['grav'] + eotvos_corr + tide_corr

# %%
# apply a lowpass filter
taps = 2*240
freq = 1./240
# we resampled to the specified sampling rate when reading the data
nyquist = sampling/2
wn = freq/nyquist       # (if that wasn't the rate to begin with)
B = firwin(taps, wn, window='blackman')  # approx equivalent to matlab fir1

dfaa = filtfilt(B, 1, dgs_data['faa'])
bfaa = filtfilt(B, 1, bgm_data['faa'])

# %%
# load satellite data for comparison
sat_path = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Thompson/TN400/sandwell_tracked.llg']))
sat_grav = np.loadtxt(sat_path[0], usecols=(3,),
                      delimiter=',', skiprows=1)

# %%
# plot this data and satellite data (trim edge effects from filtering)
plt.figure(figsize=(11, 4.8))
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         dfaa[taps:-taps//2], label='TN400 DGS')
plt.plot(bgm_data.iloc[taps:-taps//2]['date_time'],
         bfaa[taps:-taps//2], label='TN400 BGM')
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         sat_grav[taps:-taps//2], label='satellite')
plt.xlabel('Timestamp')
plt.ylabel('Free air anomaly [mGal]')
plt.legend(fontsize=8)
plt.tight_layout()
plt.show()

# %%



import re

import numpy as np
from pandas import DataFrame

# TODO gaussian filter doesn't *need* time array


def gaussian_filter(x, fl):
    """ Apply a gaussian filter to a vector.

    The filtering is done via a ring buffer; results are not identical to 
    scipy.ndimage.gaussian_filter, which is why this function exists.
    Note that the filter is not applied symmetrically and there is a shift by fl.

    :param x: time series data to be filtered.
    :type x: array_like
    :param fl: length of the filter in # of samples.
    :type fl: int

    :return: **xfilt** (*ndarray*) - filtered time series x
    """

    assert len(x) > fl, 'raw data is shorter than gaussian filter'

    coeffs = _gaussian_coeffs(fl)

    # set up ring buffer
    filt_len = len(coeffs)
    ring_buffer = np.zeros(filt_len)

    xfilt = np.zeros(len(x))

    for i in range(len(x)):
        ring_buffer = np.append(ring_buffer[1:], x[i])
        filt_vec = ring_buffer*coeffs

        xfilt[i] = sum(filt_vec)

    return xfilt


def _gaussian_coeffs(fl):
    """ Calculate a gaussian filter to convolve with things that need filtering.

    :param fl: length of the filter in # of samples
    :type fl: int

    :return: **gauss_prms** (*ndarray*) - filter coefficients
    """
    half_sample = np.floor(.5*fl)
    coeff_len = int(half_sample + 1)

    frac = 6./fl  # same as NAVO and LDEO versions, and I think this is the secret sauce?

    gauss_prms = np.zeros(coeff_len)
    m = 0
    for i in range(coeff_len, 0, -1):
        x = (half_sample - i)*frac
        x = -.5*x**2
        gauss_prms[m] = np.exp(x)
        m += 1

    # reflect
    gauss_prms = np.hstack(
        (gauss_prms[1:][::-1], gauss_prms[0], gauss_prms[1:]))

    # normalize
    gauss_prms = gauss_prms/sum(gauss_prms)

    return gauss_prms


def decode_dgs_status_bits(stat, key='status'):
    """ Decode status bits from integers in DGS gravimeter files.

    Return flags as a dict (for one integer input) or a dataframe (status df input)

    :param stat: status bit(s)
    :type stat: int or pd.DataFrame
    :param key: the DataFrame column for the status bits if input is a DataFrame
    :type key: string, optional

    :return: **stat** (*dict or pd.DataFrame*) - decoded bits
    """
    # flags from Jasmine's code
    # flags = ['clamp','unclamp','GPSsync','feedback','R1','R2','ADlock','Rcvd',\
    #        'NavMod1','NavMod2','Free','SensCom','GPStime','ADsat','Datavalid','PlatCom']
    # flags from DGS documentation?
    flags = ['clamp status', 'GPSsync', 'reserved', 'feedback', 'R1', 'R2', 'ADlock', 'ack host',
             'NavMod1', 'NavMod2', 'dgs1_trouble', 'dgs2_trouble', 'GPStime', 'ADsat', 'nemo1', 'nemo2']

    assert type(stat) is int or type(
        stat) is DataFrame, 'bad input type for status bits'
    if type(stat) is int:
        bt = format(stat, '016b')
        out = {}
        for i, f in enumerate(flags):
            out[f] = bt[i]
        return out
    else:  # dataframe, add columns
        bts = np.array([format(s, '016b') for s in stat[key]])
        for i, f in enumerate(flags):
            stat.insert(i+1, f, [b[i] for b in bts])
        return stat


def clean_ini_to_toml(ini_file):
    """ Read in a .ini file and try to rewrite as toml-compliant.

    This uses simple, prescriptive regex stuff to (hopefully) clean out
    ini conventions that don't work with toml.
    It writes out a toml file with the same name as the input, with
    the extension .ini replaced by .toml

    :param ini_file: path to input ini file

    :return: **opath** (*string*) - path to output toml file
    """

    with open(ini_file, 'r') as file:
        text = file.read()  # just read the whole thing

    # fix comments (// to #)
    text = re.sub('//', '#', text)

    # clean out tabs
    text = re.sub('\t', '  ', text)

    # fix TRUE and FALSE
    text = re.sub('TRUE', 'true', text)
    text = re.sub('FALSE', 'false', text)

    text = re.sub('$', '', text)

    # fix unquoted strings and write
    opath = ini_file.rstrip('ini')+'toml'
    fo = open(opath, 'w')
    lines = text.split('\n')
    for ln in lines:
        if ln.startswith('$'):
            ln = re.sub(r'\$', '', ln)  # clean out $ for talkers

        if ln.startswith('[') or ln.startswith('#') or ln == '':
            fo.write(ln)
            fo.write('\n')
            continue

        val = ln.split('=')[1]  # get value from key=value pair
        extra_comm = ''
        if '#' in val:
            val = val.split('#')[0]
            extra_comm = '='.join(val.split('#')[1:])
        val = val.strip()

        if val == 'true' or val == 'false':  # skip bools
            fo.write(ln)
            fo.write('\n')
            continue
        if val.startswith("\""):  # skip strings already quoted
            fo.write(ln)
            fo.write('\n')
            continue
        try:
            float(val)
        except:
            val = "\"" + val + "\""
            ln = '= '.join((ln.split('=')[0], val))
            if extra_comm != '':
                ln = '   #'.join((ln, extra_comm))
        fo.write(ln)
        fo.write('\n')

    fo.close()
    return opath


class _SnappingCursor:
    """A cross-hair cursor that snaps to the closest *x* point on a line

    This is for a two-axis figure where the first axis (ax1) 
    has a cross-hair cursor, and the second axis (ax2) has a dot that's linked
    to the cursor in ax1. Basically, we can plot two different values for one
    time series, and use a cursor over one of them to move a dot around on the other one.

    The time series with the cross hair cursor must have x values that are unique and sorted.

    Since this class is designed for a specific use case, there are some hard-coded assumptions
    about the names of columns in the DataFrame being used for plotting.

    .. The specific use case here is a plot of FAA vs time (so, x unique and sorted) with 
       the cursor, and a dot moving around on a map view of a ship track (x not necessarily
       unique or sorted).

    .. The two are linked via the pandas DataFrame "data", which has columns *tsec*,
       *lon_new*, and *lat_new* 

    The cursor can also register mouse clicks so the user can interactively select points
    that are accessible later.

    For usage, see the example script interactive_line_pick.py included with the
    shipgrav package

    Heavily adapted from the matplotlib cross-hair cursor demo at
    https://matplotlib.org/stable/gallery/event_handling/cursor_demo.html

    :param ax1: first axis, for cursor; x axis unique and sorted (tsec in data)
    :param ax2: second axis, for points; lon_new vs lat_new in data
    :param line: the line plotted in ax1, of type matplotlib.lines.Line2D
    :param dot: single point plotted in ax2 that will move around, also Line2D
    :param data: pd.DataFrame with tsec, lon_new, and lat_new columns that correspond
        to what's plotted in both axes
    """

    def __init__(self, ax1, ax2, line, dot, data):
        self.ax1 = ax1
        self.dot = dot
        self.data = data
        minx = ax1.get_xlim()[0]
        maxy = ax1.get_ylim()[1]
        self.horizontal_line = ax1.axhline(y=maxy, color='k', lw=0.8, ls='--')
        self.vertical_line = ax1.axvline(x=minx, color='k', lw=0.8, ls='--')
        self.x, self.y = line.get_data()
        self._last_index = None

        self.x_seg = []
        self.y_seg = []
        self.i_seg = []
        self.scatters = ax2.plot(
            self.x_seg, self.y_seg, marker='o', color='r', linestyle=None)

    def set_cross_hair_visible(self, visible):
        need_redraw = self.horizontal_line.get_visible() != visible
        self.horizontal_line.set_visible(visible)
        self.vertical_line.set_visible(visible)
        return need_redraw

    def on_mouse_move(self, event):
        if not event.inaxes == self.ax1:
            self._last_index = None
            need_redraw = self.set_cross_hair_visible(False)
            if need_redraw:
                self.ax1.figure.canvas.draw()
        else:
            self.set_cross_hair_visible(True)
            x, y = event.xdata, event.ydata
            index = min(np.searchsorted(self.x, x), len(self.x) - 1)
            if index == self._last_index:
                return  # still on the same data point. Nothing to do.
            self._last_index = index
            x = self.x[index]
            y = self.y[index]
            # update the line positions for the cursor
            self.horizontal_line.set_ydata([y])
            self.vertical_line.set_xdata([x])

            # and also update ax2 dot
            ind_x = np.argmin(abs(x - self.data['tsec']))
            self.dot.update({'xdata': [self.data.iloc[ind_x]['lon_new'],], 'ydata': [
                            self.data.iloc[ind_x]['lat_new'],]})
            self.ax1.figure.canvas.draw()

    def on_mouse_click(self, event):
        if event.button == 1:
            x, y = event.xdata, event.ydata
            ind_x = np.argmin(abs(x - self.data['tsec']))
            self.x_seg.append(self.data.iloc[ind_x]['lon_new'])
            self.y_seg.append(self.data.iloc[ind_x]['lat_new'])
            self.i_seg.append(ind_x)
            self.scatters[0].update({'xdata': self.x_seg, 'ydata': self.y_seg})

        if event.button == 3:  # pop the last element of the lists
            self.x_seg.pop()
            self.y_seg.pop()
            self.i_seg.pop()
            self.scatters[0].update({'xdata': self.x_seg, 'ydata': self.y_seg})

        self.ax1.figure.canvas.draw()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys, os

# extensions/moduels are in another dir; use abspath to make it work
root = os.path.abspath('../')
sys.path.insert(0, root)

# import modules for autodoc
import shipgrav
import shipgrav.io
import shipgrav.grav
import shipgrav.nav
import shipgrav.utils

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'shipgrav'
copyright = '2024, Hannah Mark'
author = 'Hannah Mark'
release = '0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']
html_sidebars = {
    '**': [
        'about.html',
        'mynavigation.html',
        'searchbox.html',
    ]
}
html_theme_options = {
    'github_user': 'PFPE',
    'github_repo': 'shipgrav',
    'description': 'Marine gravity processing for UNOLS',
    #'show_powered_by': False,
    #'page_width': '1240px',
    #'sidebar_width': '220px',
    'extra_nav_links': {'Project on Github': 'https://github.com/PFPE/shipgrav'}

}

## -- Options for LaTeX output -------------------------------------------------
#latex_additional_files = ['tex/pfpe-manual.cls']
#latex_docclass = {'manual': 'pfpe-manual'}



# %% [markdown]
# ### Example script for calculating the mantle Bouger anomaly (MBA) and residual MBA (RMBA), and estimating crustal thickness variations, using data from a Ride transit.
#
# This script works with a file from the interactive line picker script
# that is provided in a zenodo repository at https://doi.org/10.5281/zenodo.12733929
# The file is automatically downloaded by this script using pooch
#
# It is hard-coded to run on a specific example file but could be
# adjusted (see below)
#
# Tthe RMBA thermal correction is highly sensitive to edge effects,
# Here we will demo the process of embedding the line segment
# in a longer line pulled from global grids: FAA, bathymetry,
# sediment thickness, and plate age.
# For a cruise with multibeam and magnetics (for bathymetry and age)
# both of those could be embedded alongside FAA. For this cruise,
# everything but the FAA is entirely pulled from global grids.
#
# Since this is an example running on a known input file, we have pre-
# tracked the coordinates in the following global grids:
#
# #### Bathymetry:
#
# The GEBCO_2014 Grid, version 20141103, http://www.gebco.net
#
# #### FAA:
#
# Sandwell, D. T., R. D. Müller, W. H. F. Smith, E. Garcia, R. Francis (2014).
# New global marine gravity model from CryoSat-2 and Jason-1 reveals buried
# tectonic structure, Science 346(6205), pp. 65-67,
# doi: 10.1126/science.1258213
#
# #### Sediment thickness:
#
# Divins, D.L. (2003). Total Sediment Thickness of the World's Oceans &
# Marginal Seas, NOAA National Geophysical Data Center, Boulder, CO.
#
# #### Plate age:
#
# Seton, M., Müller, R. D., Zahirovic, S., Williams, S., Wright, N., Cannon, J.,
# Whittaker, J., Matthews, K., McGirr, R., (2020). A global dataset of
# present-day oceanic crustal age and seafloor spreading parameters,
# Geochemistry, Geophysics, Geosystems, doi: 10.1029/2020GC009214
#
# and
#
# Müller, R. D., Zahirovic, S., Williams, S. E., Cannon, J., Seton, M.,
# Bower, D. J., Tetley, M. G., Heine, C., Le Breton, E., Liu, S., Russell, S. H. J.,
# Yang, T., Leonard, J., and Gurnis, M. (2019). A global plate model
# including lithospheric deformation along major rifts and orogens since the
# Triassic. Tectonics, 38, doi: 10.1029/2018TC005462
# (age grids accessed from earthbyte.org)
#
# To run this example using a different input file, you will need
# to supply paths to grids of bathymetry, FAA, sediment thickness,
# and plate age; and make sure that units match what the script expects.

# %% 
import os

import matplotlib.pyplot as plt
import numpy as np
import pooch
import shipgrav.grav as sgg
from geographiclib.geodesic import Geodesic
from pandas import read_csv
from scipy.interpolate import interp1d
from scipy.signal import filtfilt, firwin
from tqdm import tqdm

# %%
# read the data
ship = 'Ride'
cruise = 'SR2312'

seg_file = dgs_files = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/gravimeter/DGS/line-segments/example_segment_s3_330234-428488.dat']))

data = read_csv(seg_file[0], parse_dates=[16,])  # fully offshore Newport, OR
# this segment is a straight-ish line; it has a lot of stations along it
# where the ship was more or less stationary but that's ok

# %%
# set up coordinates for extending the ends of the line
wgs = Geodesic.WGS84  # object for calculating line extensions etc
faz01 = wgs.Inverse(data.iloc[0]['lat_new'], data.iloc[0]['lon_new'],
                    data.iloc[-1]['lat_new'], data.iloc[-1]['lon_new'])['azi2']  # faz at end
faz10 = wgs.Inverse(data.iloc[-1]['lat_new'], data.iloc[-1]['lon_new'],
                    data.iloc[0]['lat_new'], data.iloc[0]['lon_new'])['azi2']  # faz from 0
l_seg = wgs.Inverse(data.iloc[0]['lat_new'], data.iloc[0]['lon_new'],
                    data.iloc[-1]['lat_new'], data.iloc[-1]['lon_new'])['s12']  # segment length (m)
n_seg = len(data)
l_ext = 500.e3  # default 500 km extension on each end for long lines
if l_seg < l_ext:  # but if it's a short segment, just use the segment length on each end
    l_ext = l_seg
# calculate distances btwn points to get avg spacing for extensions
xrng = np.zeros(len(data))
for i in tqdm(range(1, len(data)),desc='calculating ranges'):
    xrng[i] = wgs.Inverse(data.iloc[0]['lat_new'], data.iloc[0]['lon_new'],
                          data.iloc[i]['lat_new'], data.iloc[i]['lon_new'])['s12']
dx_avg = np.mean(np.diff(xrng))
n_ext = int(l_ext/dx_avg)

# %%
# get the actual coordinate points for each extended side
frnt_ext = np.zeros((n_ext, 2))
back_ext = np.zeros((n_ext, 2))
for i in tqdm(range(n_ext),desc='calculating extensions'):
    frnt_pt = wgs.Direct(data.iloc[0]['lat_new'],
                         data.iloc[0]['lon_new'], faz10, dx_avg*(i+1))
    back_pt = wgs.Direct(data.iloc[-1]['lat_new'],
                         data.iloc[-1]['lon_new'], faz01, dx_avg*(i+1))
    frnt_ext[i, 0] = frnt_pt['lon2']
    frnt_ext[i, 1] = frnt_pt['lat2']
    back_ext[i, 0] = back_pt['lon2']
    back_ext[i, 1] = back_pt['lat2']

# compile full coordinate lists for the extended line, write out for grdtrack
a_lon = np.hstack(
    [frnt_ext[:, 0][::-1], data['lon_new'].values, back_ext[:, 0]])
a_lat = np.hstack(
    [frnt_ext[:, 1][::-1], data['lat_new'].values, back_ext[:, 1]])

# %%
########################################################################
# this would be the point where you would track through gridfiles
# if using a different line than the example file.
########################################################################
# with open('temp.ll','w') as f:
#    for i in range(len(a_lon)):
#        f.write('%15f %15f\n' % (a_lon[i], a_lat[i]))
#
# track FAA, bathymetry, ages, sediments in global grids
# dep_grd = os.path.expanduser('~/Data/GEBCO/RN-4105_1418140094903/gebco.grd')
# faa_grd = os.path.expanduser('~/Data/other_grids/sandwell_grid/grav_32.1.nc')
# sed_grd = os.path.expanduser('~/Data/other_grids/sedthick_world.grd')
# age_grd = os.path.expanduser('~/Data/Muller_agegrids/age.2020.1.GeeK2007.1m.nc')
# os.system('gmt grdtrack temp.ll -G%s -G%s -G%s -G%s > tracked.llm' % (dep_grd, faa_grd, sed_grd, age_grd))
########################################################################

# %%
# since we've already pre-tracked the example line, we'll just load in the
# info here

track_file = dgs_files = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/gravimeter/DGS/line-segments/tracked.llm']))
track = read_csv(track_file[0], sep='\t',
                 names=['lon', 'lat', 'dep', 'faa', 'sed', 'age'])
# NOTE that if you are working with data from near a coast, line extensions
# might end up on land. Check for NaN values in your tracked file.

# %%
# filter FAA from the segment since Sandwell grid is very long-wavelength
sampling = 1
taps = 2*240
freq = 1./240
nyquist = sampling/2
wn = freq/nyquist
B = firwin(taps, wn, window='blackman')  # approx equivalent to matlab fir1
ffaa = filtfilt(B, 1, data['faa'])

# %%
# embed our FAA in the tracked line (everything else remains tracked)
track.loc[n_ext+taps:n_ext+taps+len(ffaa[taps:-taps])-1,'faa'] = ffaa[taps:-taps]
# NOTE you may want to check your embedded FAA for large jumps at the ends
# and smooth them out into the gridded data. Not bothering here because
# the jumps aren't *too* big* and also we don't care that much about the
# results of this example, but for real analyses, you should care!

# %%
# interpolate everything to an even X spacing (after recalculating total distance)
xpts_line = np.zeros(len(track))
for i in tqdm(range(1, len(track)),desc='interpolating to even spacing'):
    xpts_line[i] = wgs.Inverse(track.iloc[0]['lat'], track.iloc[0]['lon'],
                               track.iloc[i]['lat'], track.iloc[i]['lon'])['s12']
# NOTE be careful with the number of points here
xobs = np.linspace(0, xpts_line[-1], int(1e3))
# bc it determines the dx for the crustal thickness calc which has an effect on
# frequencies/wavelengths that can reasonably be used. if dx is too small zdown
# has to also be very small, which makes less sense for crustal thickness
# Basically, know that there *is* such thing as sampling too finely. 1e3 works ok here.

fgrav = interp1d(xpts_line, track['faa'].values)
FAA_int = fgrav(xobs)
fdep = interp1d(xpts_line, track['dep'].values)
dep_int = fdep(xobs)
fage = interp1d(xpts_line, track['age'].values)
age_int = fage(xobs)
fsed = interp1d(xpts_line, track['sed'].values)
sed_int = fsed(xobs)

# %%
# calculate topography, sediment, and crust corrections (for MBA)
# and thermal correction (for RMBA)
# sed density approx based on Hamilton 1978, doi: 10.1121/1.381747
rho = np.array((1000, 2100, 2900))
a = 3.e-5  # thermal expansion coeff; 3e-5 is default value for therm_Z calculation
Tm = 1350
dtemp = 50
temps = np.arange(0, Tm+1, dtemp)
rhom = 3300
rhos = rhom*(1-a*(temps-Tm))
rho = np.append(rho, rhos)
drho = np.diff(rho)
dz_c = 6000

anom_w = sgg.grav1d_padded(xobs, dep_int, 0, drho[0])  # topography
anom_s = sgg.grav1d_padded(xobs, dep_int-sed_int, 0, drho[1])  # sediment
anom_c = sgg.grav1d_padded(xobs, dep_int-sed_int-dz_c, 0, drho[2])  # crust

MBA = FAA_int - anom_w - anom_s - anom_c  # mantle Bouger anomaly

# %%
# correct for plate cooling (RMBA)
corrs = np.zeros((len(temps)-1, len(xobs)))  # calculate gravity due to temps
for i in range(len(temps)-1):  # loop over a set of isotherms defined above
    ziso, _ = sgg.therm_Z_halfspace(age_int, temps[i], time=True, Tm=Tm, a=a)
    corrs[i, :] = sgg.grav1d_padded(
        xobs, dep_int-sed_int-ziso-dz_c, 0, drho[i+3])

# %%
# RMBA is MBA minus summed thermal correction
RMBA = MBA - np.sum(corrs, axis=0)

# %%
# estimate crustal thickness
nx = len(xobs)
ny = 1  # we're only working with a line, but the function works for a 2D region
dx = np.mean(np.diff(xobs))/1e3  # spacing is even; this is in km
dy = 0
zdown = 5  # in km; might use something more like 10 for a longer line/larger dx
rho = 0.4  # density contrast
# in km; keep these 2 values small for short line (otherwise more like 100/45)
wlarge = 30
wsmall = 5

# do the calculation with back=True so it also returns the recovered
# RMBA as a check
cthick, rev_grav = sgg.crustal_thickness_2D(RMBA, nx=nx, ny=ny, dx=dx, dy=dy,
                                            zdown=zdown, rho=rho, wlarge=wlarge, wsmall=wsmall, back=True)

# %%
# make some plots to see what we got out of this
plt.figure()
plt.plot(xobs/1e3, np.real(rev_grav), label='recovered RMBA')
plt.plot(xobs/1e3, RMBA, label='RMBA')
plt.xlabel('distance along line [km]')
plt.ylabel('anomaly [mGal]')
plt.legend(fontsize=8)

cutoff = 100  # index for trimming edge effects from the crustal thickness calc; somewhat arbitrary
plt.figure()
plt.plot(xobs[cutoff:-cutoff]/1e3, np.real(cthick[cutoff:-cutoff]))
plt.xlabel('distance along line [km]')
plt.ylabel('crustal thickness variation [km]')

plt.show()

# %%



"""
Tests for shipgrav.utils
"""
import unittest

import numpy as np
import shipgrav.utils as sgu


class utilsTestCase(unittest.TestCase):
    def test_gaussian_filter(self):
        # make a spike, filtfilt for no shift, check amplitude
        test = np.zeros(101)
        test[50] = 1
        test_fl = sgu.gaussian_filter(test, 10)
        test_ffll = sgu.gaussian_filter(test_fl[::-1], 10)
        self.assertTrue(test_ffll[50] - 0.12975 < 0.001)

    def test_status_decode(self):
        # make a code, decode it, check a few of the 16 status bits
        stat = 12345
        decoded = sgu.decode_dgs_status_bits(stat)
        self.assertEqual(decoded['clamp status'], '0')
        self.assertEqual(decoded['GPSsync'], '0')
        self.assertEqual(decoded['GPStime'], '1')
        self.assertEqual(decoded['feedback'], '1')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(utilsTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# %% [markdown]
# ### Example script for reading DGS laptop data and MRUs from an R/V Ride transit, and calculating coherence between MRUs and monitors
#
# Data files are downloaded by the script using pooch
#
# Read DGS laptop and navigation files
#
# Correct for meter bias with info from shipgrav
#
# Use timestamps to sync more accurate nav with the gravity data.
#
# Calculate FAA (free air anomaly) for laptop data
#
# Read MRUs (pitch/roll/heave) for all pairs of monitors and MRUs, interpolate MRU to grav sample rate and calculate coherence between monitor and 1000-pt moving average of MRU
#
# (and plot coherence)

# %%
import os
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pooch
import tomli as tm
from pandas import concat, to_datetime
from scipy.interpolate import interp1d
from scipy.signal import coherence, filtfilt, firwin
from tqdm import tqdm

import shipgrav.grav as sgg
import shipgrav.io as sgi

# %%
# set some general metadata
ship = 'Ride'
cruise = 'SR2312'       # this is used for filepaths
sampling = 1            # 1 Hz - data should be at this rate already

# read a few constants etc from our toml database file
with open('../shipgrav/database.toml', 'rb') as f:
    info = tm.load(f)
nav_tag = info['nav-talkers'][ship]
biases = info['bias-values'][ship]

# %%
# get the DGS and nav data from R2R
dgs_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/SR2312/fileset/157179", 
        known_hash="53f53c45aa59ce19cd1e75e3d847b5697123ad0a1296aa2be28bf26ff0ad19ac", progressbar=True, 
        processor=pooch.Untar(
            members=['SR2312/157179/data/AT1M-25_20230616_167.dat',
                     'SR2312/157179/data/AT1M-25_20230617_168.dat',
                     'SR2312/157179/data/AT1M-25_20230618_169.dat',
                     'SR2312/157179/data/AT1M-25_20230619_170.dat',
                     'SR2312/157179/data/AT1M-25_20230620_171.dat']))

nav_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/SR2312/fileset/157188", 
        known_hash="eb5992eadd87b09f66308a4d577c6ba6965d8ed8489959ffaed8bf5556ee712f", progressbar=True, 
        processor=pooch.Untar(
            members=['SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-16.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-17.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-18.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-19.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-20.txt']))

# %%
# read and sort the nav data
gps_nav = sgi.read_nav(ship, nav_files, talker='GPGGA')
gps_nav.sort_values('time_sec', inplace=True)
gps_nav.reset_index(inplace=True, drop=True)

# %%
# read and sort the DGS laptop data
dgs_data = sgi.read_dgs_laptop(dgs_files, ship)
dgs_data.sort_values('date_time', inplace=True)
dgs_data.reset_index(inplace=True, drop=True)
dgs_data['tsec'] = [e.timestamp()
                    for e in dgs_data['date_time']]  # get posix timestamps
dgs_data['grav'] = dgs_data['rgrav'] + biases['dgs']

# %%
# sync data geographic coordinates to nav by interpolating with timestamps
# (interpolators use posix timestamps, not datetimes)
gps_lon_int = interp1d(gps_nav['time_sec'].values, gps_nav['lon'].values,
                       kind='linear', fill_value='extrapolate')
gps_lat_int = interp1d(gps_nav['time_sec'].values, gps_nav['lat'].values,
                       kind='linear', fill_value='extrapolate')
dgs_data['lon_new'] = gps_lon_int(dgs_data['tsec'].values)
dgs_data['lat_new'] = gps_lat_int(dgs_data['tsec'].values)

# %%
# calculate corrections for FAA
ellipsoid_ht = np.zeros(len(dgs_data))  # we are working at sea level
lat_corr = sgg.wgs_grav(dgs_data['lat_new']) + \
    sgg.free_air_second_order(dgs_data['lat_new'], ellipsoid_ht)
eotvos_corr = sgg.eotvos_full(dgs_data['lon_new'].values, dgs_data['lat_new'].values,
                              ellipsoid_ht, sampling)
tide_corr = sgg.longman_tide_prediction(
    dgs_data['lon_new'], dgs_data['lat_new'], dgs_data['date_time'])

dgs_data['faa'] = dgs_data['grav'] - lat_corr + eotvos_corr + tide_corr
dgs_data['full_field'] = dgs_data['grav'] + eotvos_corr + tide_corr

# %%
# apply a lowpass filter to FAA
taps = 2*240
freq = 1./240
# we resampled to the specified sampling rate when reading the data
nyquist = sampling/2
wn = freq/nyquist       # (if that wasn't the rate to begin with)
B = firwin(taps, wn, window='blackman')  # approx equivalent to matlab fir1

ffaa = filtfilt(B, 1, dgs_data['faa'])

# %%
# get some MRU time series files from zenodo
# read in some other time series from MRUs
mru_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/SR2312/fileset/157177", 
        known_hash="ca12af51a33983f77d775098c06744e509baf31531a2afc37e39f1775f51d7f8", progressbar=True, 
        processor=pooch.Untar(
            members=['SR2312/157177/data/SR2312_mru_hydrins_navbho-2023-06-16.txt',
                     'SR2312/157177/data/SR2312_mru_hydrins_navbho-2023-06-17.txt',
                     'SR2312/157177/data/SR2312_mru_hydrins_navbho-2023-06-18.txt',
                     'SR2312/157177/data/SR2312_mru_hydrins_navbho-2023-06-19.txt',
                     'SR2312/157177/data/SR2312_mru_hydrins_navbho-2023-06-20.txt']))
yaml_file = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/openrvdas/doc/devices/IXBlue.yaml']))

# %%
talk = 'PASHR'
dats = []
# note that the MRU data will need to be sorted in the same time series order as the DGS data
# sorting by filename accomplishes this since the paths have dates
for mf in tqdm(np.sort(mru_files),desc='reading MRUs'):
    dat, col_info = sgi.read_other_stuff(yaml_file[0], mf, talk)
    dats.append(dat)
mru_dat = concat(dats, ignore_index=True)
del dats, dat  # cleanup

# %%
# we have some prior knowledge about this data that lets us find the timestamps:
mru_dat['date_time'] = to_datetime(mru_dat['mystery'], utc=True)
mru_dat.drop('mystery', axis=1, inplace=True)
mru_dat['tsec'] = [e.timestamp() for e in mru_dat['date_time']]

# %%
# if we want to look at coherence between monitors and MRUs, have to interpolate first
# because monitors are at 2Hz
for motion in ['Pitch', 'Roll', 'Heave']:
    plt.figure()
    plt.title('%s coherence with monitors' % motion)
    # interpolate a sort of max envelope value for each MRU to the grvimeter sample rate
    mru_interp = np.interp(dgs_data.tsec, mru_dat.tsec,
                           mru_dat['%s:g' % motion].rolling(1000).max())
    mru_interp[np.isnan(mru_interp)] = 0  # Nans are no good for coherence
    for monitor in ['vcc', 've', 'al', 'ax']:
        freq, coh = coherence(dgs_data[monitor], mru_interp)
        plt.semilogy(freq, coh, label=monitor)
    plt.legend(fontsize=8)
    plt.xlabel('Frequency [Hz]')

plt.show()

# %%



"""
Tests for shipgrav.grav
"""
import unittest

import numpy as np
import shipgrav.grav as sgg


class gravNoDataTestCase(unittest.TestCase):
    def test_grav1d(self):
        # arbitrary numbers here, not real
        test_grav = sgg.grav1d_padded(np.linspace(
            0, 100, 10), np.linspace(10, 15, 10), 1, 0.4)
        self.assertTrue(test_grav[1] - 0.000157 < 0.001)

    def test_halfspace_T(self):
        T, W = sgg.therm_halfspace(np.array([20e3]), np.array([1e3]), u=0.02)
        self.assertTrue(T[0] - 135.2263 < 0.001)
        self.assertTrue(W[0] - 368.3393 < 0.001)

    def test_halfspace_Z(self):
        Z, W = sgg.therm_Z_halfspace(np.array([20e3]), 135.2263, u=0.02)
        self.assertTrue(Z[0] - 1e3 < 0.001)
        self.assertTrue(W[0] - 368.3393 < 0.001)

    def test_plate_T(self):
        T, W = sgg.therm_plate(np.array([20e3]), np.array([1e3]), u=0.02)
        self.assertTrue(T[0] - 134.0867 < 0.001)
        self.assertTrue(W[0] - 381.6513 < 0.001)

    def test_plate_Z(self):
        Z = sgg.therm_Z_plate(
            np.array([20e3]), np.array([134.0867]), u=0.02)
        self.assertEqual(Z[0], 1000.)

    def test_crustalthickness(self):
        rng = np.random.default_rng(123)  # seeded
        signal = 10*rng.random(1000)
        C = sgg.crustal_thickness_2D(signal)
        self.assertTrue(np.real(C[0])[0] - 0.04019 < 0.001)

        C2 = sgg.crustal_thickness_2D(signal, back=True)
        self.assertEqual(np.real(C[0])[0], np.real(C2[0][-1])[0])

    def test_grav2d_folding(self):
        rng = np.random.default_rng(123)  # seeded
        X = np.arange(5); Y = np.arange(5)
        Z = rng.random((5,5))
        sdat = sgg.grav2d_folding(X, Y, Z, 100, 100,  drho=0.6, dz=6000, ifold=True, npower=5)

        self.assertTrue(sdat[0,0] + 0.0043244755 < 0.001)

    def test_grav2d_layer(self):
        rng = np.random.default_rng(123)  # seeded
        rho = 1e4*rng.random((5,5))
        sdat = sgg.grav2d_layer_variable_density(rho, 100, 100, 3, 5)
        self.assertTrue(sdat[0,0] - 130.877533 < 0.001)

def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(gravNoDataTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# %% [markdown]
# ### Example script for reading and calibrating DGS raw (aka serial) data from an R/V Ride transit, with comparison to laptop (lightly processed) files output from the meter.
#
# Data files are downloaded by the script using pooch
#
# Read DGS (serial and laptop) and navigation files
#
# Calibrate the raw counts
#
# Correct for meter bias with info from shipgrav
#
# Use timestamps to sync more accurate nav with the gravity data.
#
# Plot laptop and raw data to compare
#
# Calculate FAA (free air anomaly) for laptop data
#
# Plot laptop FAA along with satellite
#
# satellite data was tracked from v32.1 Global Gravity grid, which
# includes data from SIO, NOAA, and NGA.
#
# Reference: Sandwell et al. (2014) New global marine gravity model
# from CryoSat-2 and Jason-1 reveals buried tectonic struture.
# Science 346(6205), DOI: 10.1126/science.1258213

# %%
import os
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pooch
import tomli as tm
from scipy.interpolate import interp1d
from scipy.signal import filtfilt, firwin
from tqdm import tqdm

import shipgrav.grav as sgg
import shipgrav.io as sgi
import shipgrav.utils as sgu

# %%
# set some general metadata
ship = 'Ride'
cruise = 'SR2312'       # this is used for filepaths
sampling = 1            # 1 Hz - data should be at this rate already

# read a few constants etc from our toml database file
with open('../shipgrav/database.toml', 'rb') as f:
    info = tm.load(f)
nav_tag = info['nav-talkers'][ship]
biases = info['bias-values'][ship]
cal_factor = info['dgs-stuff']['calibration_factor']

# %%
# get the data: raw DGS from zenodo, laptop and nav from R2R
dgs_files_raw = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/gravimeter/DGS/serial/SR2312_grav_dgs-2023-06-16.txt',
                    'data/Ride/SR2312/gravimeter/DGS/serial/SR2312_grav_dgs-2023-06-17.txt',
                    'data/Ride/SR2312/gravimeter/DGS/serial/SR2312_grav_dgs-2023-06-18.txt',
                    'data/Ride/SR2312/gravimeter/DGS/serial/SR2312_grav_dgs-2023-06-19.txt',
                    'data/Ride/SR2312/gravimeter/DGS/serial/SR2312_grav_dgs-2023-06-20.txt']))

dgs_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/SR2312/fileset/157179", 
        known_hash="53f53c45aa59ce19cd1e75e3d847b5697123ad0a1296aa2be28bf26ff0ad19ac", progressbar=True, 
        processor=pooch.Untar(
            members=['SR2312/157179/data/AT1M-25_20230616_167.dat',
                     'SR2312/157179/data/AT1M-25_20230617_168.dat',
                     'SR2312/157179/data/AT1M-25_20230618_169.dat',
                     'SR2312/157179/data/AT1M-25_20230619_170.dat',
                     'SR2312/157179/data/AT1M-25_20230620_171.dat']))

nav_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/SR2312/fileset/157188", 
        known_hash="eb5992eadd87b09f66308a4d577c6ba6965d8ed8489959ffaed8bf5556ee712f", progressbar=True, 
        processor=pooch.Untar(
            members=['SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-16.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-17.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-18.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-19.txt',
                     'SR2312/157188/data/SR2312_mru_seapath330_navbho-2023-06-20.txt']))

# %%
# read and sort the nav data
gps_nav = sgi.read_nav(ship, nav_files, talker='GPGGA')
gps_nav.sort_values('time_sec', inplace=True)
gps_nav.reset_index(inplace=True, drop=True)

# %%
# read and sort the DGS laptop data
dgs_data = sgi.read_dgs_laptop(dgs_files, ship)
dgs_data.sort_values('date_time', inplace=True)
dgs_data.reset_index(inplace=True, drop=True)
dgs_data['tsec'] = [e.timestamp()
                    for e in dgs_data['date_time']]  # get posix timestamps
dgs_data['grav'] = dgs_data['rgrav'] + biases['dgs']

# %%
# sync data geographic coordinates to nav by interpolating with timestamps
# (interpolators use posix timestamps, not datetimes)
gps_lon_int = interp1d(gps_nav['time_sec'].values, gps_nav['lon'].values,
                       kind='linear', fill_value='extrapolate')
gps_lat_int = interp1d(gps_nav['time_sec'].values, gps_nav['lat'].values,
                       kind='linear', fill_value='extrapolate')
dgs_data['lon_new'] = gps_lon_int(dgs_data['tsec'].values)
dgs_data['lat_new'] = gps_lat_int(dgs_data['tsec'].values)

# %%
# calculate corrections for FAA
ellipsoid_ht = np.zeros(len(dgs_data))  # we are working at sea level
lat_corr = sgg.wgs_grav(dgs_data['lat_new']) + \
    sgg.free_air_second_order(dgs_data['lat_new'], ellipsoid_ht)
eotvos_corr = sgg.eotvos_full(dgs_data['lon_new'].values, dgs_data['lat_new'].values,
                              ellipsoid_ht, sampling)
tide_corr = sgg.longman_tide_prediction(
    dgs_data['lon_new'], dgs_data['lat_new'], dgs_data['date_time'])

dgs_data['faa'] = dgs_data['grav'] - lat_corr + eotvos_corr + tide_corr
dgs_data['full_field'] = dgs_data['grav'] + eotvos_corr + tide_corr

# %%
# the serial files for this cruise have some issues, so clean them up first
dgs_raw_clean = []
for path in tqdm(dgs_files_raw,desc='cleaning serial files'):  # clean up weird Ride serial files
    splitpath = path.split('/')
    newname = 'clean_' + splitpath[-1]
    splitpath[-1] = newname
    opath = '/'.join(splitpath)
    dgs_raw_clean.append(opath)
    if os.path.isfile(opath):
        continue
    with open(path, 'r') as f:
        A = f.readlines()
    long = []
    for line in A:
        if 'AT1M' in line and len(line.split(',')) == 19:
            long.append(line)
    with open(opath, 'w') as f:
        for line in long:
            f.write(line)
dgs_files_raw = np.sort(dgs_raw_clean)

# %%
# read the raw (serial) DGS data
raw_dgs = sgi.read_dgs_raw(dgs_files_raw, ship)

# %%
# read some meter info for the raw data calibration
# find the meter config file
ini_file = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/Meter.ini']))
opath = sgu.clean_ini_to_toml(ini_file[0])  # make it a little easier to read
meter_conf = tm.load(open(opath, 'rb'))

# %%
# calibrate stuff (though we won't use all of these things here)
# 24-bit channels
raw_dgs['grav_cal'] = (raw_dgs['Gravity']*meter_conf['Sensor']['GravCal']/cal_factor) + \
    meter_conf['Sensor']['g0']
raw_dgs['long_cal'] = (raw_dgs['Long']*meter_conf['Sensor']['LongCal']/cal_factor) + \
    meter_conf['Sensor']['LongOffset']
raw_dgs['cross_cal'] = (raw_dgs['Cross']*meter_conf['Sensor']['CrossCal']/cal_factor) + \
    meter_conf['Sensor']['CrossOffset']
raw_dgs['beam_cal'] = (raw_dgs['Beam']*meter_conf['Sensor']['beamgain']/cal_factor) + \
    meter_conf['Sensor']['beamzero']
# 10-bit channels
raw_dgs['temp_cal'] = (raw_dgs['Temp']*meter_conf['Sensor']['stempgain']) + \
    meter_conf['Sensor']['stempoffset']
raw_dgs['pressure_cal'] = (raw_dgs['Pressure']*meter_conf['Sensor']['pressgain']) + \
    meter_conf['Sensor']['presszero']
raw_dgs['electemp_cal'] = (raw_dgs['ElecTemp']*meter_conf['Sensor']['Etempgain']) + \
    meter_conf['Sensor']['Etempzero']
raw_dgs['grav'] = raw_dgs['grav_cal'] + biases['dgs']

# %%
# plot a comparison between serial and laptop data (raw, not FAA)
# select a small portion of the data bc otherwise it's hard to see anything
d0 = dgs_data.iloc[0]['date_time']
d1 = dgs_data.iloc[500]['date_time']
plt.figure(figsize=(11, 4.8))
sl0 = dgs_data.date_time > d0
sl1 = dgs_data.date_time < d1
plt.plot(dgs_data[(sl0) & (sl1)]['date_time'],
         dgs_data[(sl0) & (sl1)]['rgrav'], color='k', label='laptop')
sl0 = raw_dgs.date_time > d0
sl1 = raw_dgs.date_time < d1
plt.plot(raw_dgs[(sl0) & (sl1)]['date_time'], raw_dgs[(sl0) &
         (sl1)]['grav_cal'], color='r', lw=.5, label='serial')
plt.xlabel('Timestamp')
plt.ylabel('Raw gravity')
plt.legend(fontsize=8)
plt.tight_layout()

# %%
# get position information for raw data points from gps nav (serial files for
# this cruise, at least, have zeros where the coordinates are expected to be)
raw_dgs['tsec'] = [e.timestamp() for e in raw_dgs['date_time']
                   ]  # posix stamps in raw file
# have some dropouts
raw_dgs['lon_new'] = gps_lon_int(raw_dgs['tsec'].values)
raw_dgs['lat_new'] = gps_lat_int(raw_dgs['tsec'].values)

# %%
# NOTE that because of issues with the serial output for this cruise, the raw data
# is not sampled at an even time interval.
plt.figure()
plt.scatter(raw_dgs.date_time[1:], np.diff(raw_dgs.tsec), marker='.')
plt.xlabel('timestamp')
plt.ylabel('sampling interval in seconds')
plt.title('uneven sampling interval for serial data files')

# %% [markdown]
# We can get location info by interpolating from the nav files, but we would need to
# interpolate the data as well to obtain (and filter) the FAA from the serial files.
# Having verified that the laptop and serial files match well where there *is* data,
# I'm not going to bother with that here.

# %%
# apply a lowpass filter to FAA
taps = 2*240
freq = 1./240
# we resampled to the specified sampling rate when reading the data
nyquist = sampling/2
wn = freq/nyquist       # (if that wasn't the rate to begin with)
B = firwin(taps, wn, window='blackman')  # approx equivalent to matlab fir1

ffaa = filtfilt(B, 1, dgs_data['faa'])

# %%
# load satellite data for comparison
sat_path = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Ride/SR2312/sandwell_tracked.llg']))
sat_grav = np.loadtxt(sat_path[0], usecols=(2,))

# %%
# plot laptop FAA and satellite data (trim edge effects from filtering)
plt.figure(figsize=(11, 4.8))
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         ffaa[taps:-taps//2], label='SR2312')
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         sat_grav[taps:-taps//2], label='satellite')
plt.xlabel('Timestamp')
plt.ylabel('Free air anomaly [mGal]')
plt.legend(fontsize=8)
plt.tight_layout()

# show all of the figures
plt.show()

# %%



# %% [markdown]
# ### Example script using data from TN400 to illustrate how cross-coupling coefficients are calculated and applied.
#
# Data files are downloaded by the script using pooch
#
# Read DGS data files
# (not bothering with navigation syncing; see dgs_bgm_comp.py)
#
# Calculate FAA
#
# Calculate various kinematic parameters
#
# Fit for cross-coupling coefficients
#
# Correct for cross-coupling
#
# Plot with and without corrections

# %%
import os
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pooch
import shipgrav.grav as sgg
import shipgrav.io as sgi
import shipgrav.nav as sgn
import tomli as tm
from scipy.interpolate import interp1d
from scipy.signal import filtfilt, firwin

# %%
# set some general metadata
ship = 'Thompson'
cruise = 'TN400'
sampling = 1
# read a few constants etc from our toml database file
with open('../shipgrav/database.toml', 'rb') as f:
    info = tm.load(f)
# nav_tag = info['nav-talkers'][ship]  # could use this for nav file read
bias_dgs = info['bias-values'][ship]['dgs']

# %%
# get the files from Zenodo: TN400 DGS laptop data
# the archive download includes some things not used for this example; those are not unpacked
dgs_files = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Thompson/TN400/gravimeter/DGS/AT1M-Grav-PROC_20220314-000001.Raw',
                    'data/Thompson/TN400/gravimeter/DGS/AT1M-Grav-PROC_20220313-000001.Raw']))

nav_files = pooch.retrieve(url="https://service.rvdata.us/data/cruise/TN400/fileset/151457",
        known_hash="76e66365c41d393510bb7ab9637305296282e9041415c1343faa171af28abf85",progressbar=True,
        processor=pooch.Untar(
            members=['TN400/151457/data/POSMV-V5-INGGA-RAW_20220313-000001.Raw',
                    'TN400/151457/data/POSMV-V5-INGGA-RAW_20220314-000001.Raw']))

# %%
# read and sort the nav data
gps_nav = sgi.read_nav(ship, nav_files)
gps_nav.sort_values('time_sec', inplace=True)
gps_nav.reset_index(inplace=True, drop=True)

# %%
# we happen to know that there are some weird nav dropouts in this dataset
# so clean them up here
bad_inds = np.where(np.diff(gps_nav['lon']) > 1)[0]
gps_nav.drop(bad_inds, axis=0, inplace=True)

# %%
# read and sort the gravimeter data
dgs_data = sgi.read_dgs_laptop(dgs_files, ship)
dgs_data.sort_values('date_time', inplace=True)
dgs_data.reset_index(inplace=True, drop=True)
dgs_data['tsec'] = [e.timestamp()
                    for e in dgs_data['date_time']]  # get posix timestamps
dgs_data['grav'] = dgs_data['rgrav'] + bias_dgs

# %%
# sync data geographic coordinates to nav by interpolating with timestamps
# (interpolators use posix timestamps, not datetimes)
gps_lon_int = interp1d(gps_nav['time_sec'].values, gps_nav['lon'].values,
                       kind='linear', fill_value='extrapolate')
gps_lat_int = interp1d(gps_nav['time_sec'].values, gps_nav['lat'].values,
                       kind='linear', fill_value='extrapolate')
dgs_data['lon_new'] = gps_lon_int(dgs_data['tsec'].values)
dgs_data['lat_new'] = gps_lat_int(dgs_data['tsec'].values)


# %%
# calculate corrections for FAA
ellipsoid_ht = np.zeros(len(dgs_data))  # we are working at sea level
lat_corr = sgg.wgs_grav(dgs_data['lat_new']) + \
    sgg.free_air_second_order(dgs_data['lat_new'], ellipsoid_ht)
eotvos_corr = sgg.eotvos_full(dgs_data['lon_new'].values, dgs_data['lat_new'].values,
                              ellipsoid_ht, sampling)
tide_corr = sgg.longman_tide_prediction(
    dgs_data['lon_new'], dgs_data['lat_new'], dgs_data['date_time'])

dgs_data['faa'] = dgs_data['grav'] - lat_corr + eotvos_corr + tide_corr
dgs_data['full_field'] = dgs_data['grav'] + eotvos_corr + tide_corr

# %%
# calculate kinematic variables and corrections for tilt correction
# (maybe not strictly necessary? depends who you ask)
gps_vn, gps_ve = sgn.latlon_to_EN(
    dgs_data['lon_new'].values, dgs_data['lat_new'].values)
gps_vn = sampling*gps_vn
gps_ve = sampling*gps_ve

gps_eacc = 1e5*sampling*np.convolve(gps_ve, sgn.tay10, 'same')
gps_nacc = 1e5*sampling*np.convolve(gps_vn, sgn.tay10, 'same')

crse, vel = sgn.ENvel_to_course_heading(gps_ve, gps_vn)
crse[np.isnan(crse)] = 0
vel[np.isnan(vel)] = 0

acc_cross, acc_long = sgn.rotate_acceleration_EN_to_cl(
    crse, gps_eacc, gps_nacc)  # gps-derived cross and long accel

# %%
# tilt correction
up_vecs = sgg.up_vecs(1/sampling, lat_corr, acc_cross,
                      acc_long, 0, 240, 0.7071, 240, 0.7071)
igf_in_plat = lat_corr*up_vecs[2, :]  # latitude correction on platform
cross_in_plat = acc_cross*up_vecs[1, :]
long_in_plat = acc_long*up_vecs[0, :]
level_error = lat_corr - igf_in_plat - long_in_plat + cross_in_plat

dgs_data = dgs_data.iloc[2:]  # the first two points' FAA has some edge effect

# %%
# calculate cross-coupling coefficients
_, model = sgg.calc_cross_coupling_coefficients(dgs_data['faa'].values, dgs_data['vcc'].values, dgs_data['ve'].values,
                                                dgs_data['al'].values, dgs_data['ax'].values, level_error.values)

print('cross coupling parameters from fit:')
print('ve %.3f, vcc %.3f, al %.3f, ax %.3f, lev %.3f' % (model.params.ve,
      model.params.vcc, model.params.al, model.params.ax, model.params.lev))

# %%
# apply cross-coupling correction and plot the (filtered) FAA
dgs_data['faa_ccp'] = dgs_data['faa'] + model.params.ve*dgs_data['ve'] + \
    model.params.vcc*dgs_data['vcc'] + \
    model.params.al*dgs_data['al'] + \
    model.params.ax*dgs_data['ax']

# filter FAA
taps = 2*240
freq = 1./240
# we resampled to the specified sampling rate when reading the data
nyquist = sampling/2
wn = freq/nyquist       # (if that wasn't the rate to begin with)
B = firwin(taps, wn, window='blackman')  # approx equivalent to matlab fir1

ffaa = filtfilt(B, 1, dgs_data['faa'])
cfaa = filtfilt(B, 1, dgs_data['faa_ccp'])

# %%
# load satellite data for comparison
sat_path = pooch.retrieve(url="https://zenodo.org/records/12733929/files/data.zip", 
        known_hash="md5:83b0411926c0fef9d7ccb2515bb27cc0", progressbar=True, 
        processor=pooch.Unzip(
            members=['data/Thompson/TN400/sandwell_tracked.llg']))
sat_grav = np.loadtxt(sat_path[0], usecols=(3,),
                      delimiter=',', skiprows=1)
sat_grav = sat_grav[2:]


# %%
plt.figure(figsize=(11, 4.8))
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         ffaa[taps:-taps//2], label='no ccp')
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         cfaa[taps:-taps//2], label='with ccp')
plt.plot(dgs_data.iloc[taps:-taps//2]['date_time'],
         sat_grav[taps:-taps//2], label='satellite')
plt.xlabel('Timestamp')
plt.ylabel('Free air anomaly [mGal]')
plt.legend(fontsize=8)
plt.tight_layout()
plt.show()

# %%



"""
Tests for shipgrav.io
This does not test every option for read (because there are many ships)
but for each overall read function it reads a snippet of an example file
and checks that the values in key columns are correct
"""
import os
import unittest

import shipgrav.io as sgi


class ioTestCase(unittest.TestCase):
    @property
    def ex_files(self):
        return __file__.rsplit(os.sep, 1)[0] + os.sep + "ex_files"

    def test_read_nav_Thompson(self):
        nav = sgi.read_nav('Thompson', f'{self.ex_files}'+os.sep+'TN400_nav.Raw', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1647129603)
        self.assertTrue(nav.iloc[0].lon + 118.6524 < 0.001)

    def test_read_nav_Atlantis(self):
        nav = sgi.read_nav('Atlantis', f'{self.ex_files}'+os.sep+'AT01_nav.gps', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1656677853)
        self.assertTrue(nav.iloc[0].lon + 70.67185265 < 0.001)

    def test_read_nav_Langseth(self):
        nav = sgi.read_nav('Langseth', f'{self.ex_files}'+os.sep+'MGL2003_nav.y2020d244', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1598832000.6212)
        self.assertTrue(nav.iloc[0].lon + 132.620965893 < 0.001)

    def test_read_nav_Revelle(self):
        nav = sgi.read_nav('Revelle', f'{self.ex_files}'+os.sep+'RR2212_nav.txt', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1667606400)
        self.assertTrue(nav.iloc[0].lon + 119.3889582 < 0.001)

    def test_read_nav_Ride(self):
        nav = sgi.read_nav('Ride', f'{self.ex_files}'+os.sep+'SR2302_nav.raw',talker='INGGA', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1674228947.583)
        self.assertTrue(nav.iloc[0].lon + 117.23672175 < 0.001)

    def test_read_nav_NBP(self):
        nav = sgi.read_nav('NBP', f'{self.ex_files}'+os.sep+'NBP_2301_nav.d013', progressbar=False)
        self.assertEqual(nav.iloc[0].time_sec, 1673568001.81)
        self.assertTrue(nav.iloc[0].lon + 179.7605851 < 0.001)

    def test_read_nav_nope(self):
        nav = sgi.read_nav('Titanic', f'{self.ex_files}'+os.sep+'TN400_nav.Raw', progressbar=False)
        self.assertEqual(nav, -999)

    def test_read_bgm_rgs(self):
        bgm = sgi.read_bgm_rgs(f'{self.ex_files}'+os.sep+'AT05_01_bgm.RGS', 'Atlantis', progressbar=False)
        self.assertEqual(bgm.iloc[0]['date_time'].timestamp(), 1656633600.445)
        self.assertEqual(bgm.iloc[0]['grav'], 980329.272)

    def test_read_bgm_raw_Atlantis(self):
        bgm = sgi.read_bgm_raw(f'{self.ex_files}'+os.sep+'AT01_bgm.BGM', 'Atlantis', progressbar=False)
        self.assertEqual(bgm.iloc[0]['date_time'].timestamp(), 1656677853.492)
        self.assertEqual(bgm.iloc[0]['counts'], 24963)
        self.assertEqual(bgm.iloc[0]['rgrav'], 124666.983189576)

    def test_read_bgm_raw_Thompson(self):
        bgm = sgi.read_bgm_raw(f'{self.ex_files}'+os.sep+'TN400_bgm.Raw', 'Thompson', progressbar=False)
        self.assertEqual(bgm.iloc[0]['date_time'].timestamp(), 1647129602.449)
        self.assertEqual(bgm.iloc[0]['counts'], 25529)
        self.assertEqual(bgm.iloc[0]['rgrav'], 127730.60800402702)

    def test_read_bgm_raw_Langseth(self):
        bgm = sgi.read_bgm_raw(f'{self.ex_files}'+os.sep+'MGL2003_bgm.y2020d244', 'Langseth', progressbar=False)
        self.assertEqual(bgm.iloc[0]['date_time'].timestamp(), 1598832000.3244)
        self.assertEqual(bgm.iloc[0]['counts'], 25229)
        self.assertEqual(bgm.iloc[0]['rgrav'], 126145.0)

    def test_read_bgm_raw_Revelle(self):
        bgm = sgi.read_bgm_raw(f'{self.ex_files}'+os.sep+'RR2212_bgm.txt', 'Revelle', progressbar=False)
        self.assertEqual(bgm.iloc[0]['date_time'].timestamp(), 1667841225.329611)
        self.assertEqual(bgm.iloc[0]['counts'], 24882)
        self.assertEqual(bgm.iloc[0]['rgrav'], 124405.2114591)

    def test_read_bgm_raw_nope(self):
        bgm = sgi.read_bgm_raw(f'{self.ex_files}'+os.sep+'TN400_bgm.Raw', 'Boaty McBoatface', progressbar=False)
        self.assertEqual(bgm, -999)

    def test_read_dgs_dat_general(self):
        dgs = sgi.read_dgs_laptop(f'{self.ex_files}'+os.sep+'DGStest_laptop.dat', 'DGStest', progressbar=False)
        self.assertEqual(dgs.iloc[0]['date_time'].timestamp(), 1562803200.0)
        self.assertEqual(dgs.iloc[0]['ve'], 0.81098)
        self.assertTrue(dgs.iloc[0]['rgrav'] - 12295.691114 < 0.0001)

    def test_read_dgs_dat_Thompson(self):
        dgs = sgi.read_dgs_laptop(f'{self.ex_files}'+os.sep+'TN400_dgs_proc.Raw', 'Thompson', progressbar=False)
        self.assertEqual(dgs.iloc[0]['date_time'].timestamp(), 1647101047.033)
        self.assertEqual(dgs.iloc[0]['ve'], 2e-5)
        self.assertTrue(dgs.iloc[0]['rgrav'] - 9995.95186 < 0.0001)

    def test_read_dgs_dat_Langseth(self):
        dgs = sgi.read_dgs_laptop(f'{self.ex_files}'+os.sep+'MGL_dgsdata.y2025d254', 'Langseth', progressbar=False)
        self.assertEqual(dgs.iloc[0]['date_time'].timestamp(), 1757548800.91)
        self.assertEqual(dgs.iloc[0]['ve'], 0.32308)
        self.assertTrue(dgs.iloc[0]['rgrav'] - 1483.1871291 < 0.0001)

    def test_read_dgs_dat_Sikuliaq(self):
        dgs = sgi.read_dgs_laptop(f'{self.ex_files}'+os.sep+'SKQ_dgs_proc.20241213T0000Z', 'Sikuliaq', progressbar=False)
        self.assertEqual(dgs.iloc[0]['date_time'].timestamp(), 1734047995.0)
        self.assertEqual(dgs.iloc[0]['ve'], 0.00235)
        self.assertTrue(dgs.iloc[0]['rgrav'] - 9288.4046923 < 0.0001)

    def test_read_dgs_dat_nope(self):
        dgs = sgi.read_dgs_laptop(f'{self.ex_files}'+os.sep+'DGStest_laptop.dat', 'Katama', progressbar=False)
        self.assertEqual(dgs, -999)

    def test_read_dgs_raw_general(self):
        dgs = sgi.read_dgs_raw(f'{self.ex_files}'+os.sep+'SR2312_dgs_raw.txt', 'Ride', progressbar=False)
        self.assertEqual(
            dgs.iloc[0]['date_time'].timestamp(), 1686873600.857719)
        self.assertEqual(dgs.iloc[0]['Gravity'], -218747)
        self.assertTrue(dgs.iloc[0]['vcc'] - 76.8771 < 0.0001)

    def test_read_dgs_raw_Thompson(self):
        dgs = sgi.read_dgs_raw(f'{self.ex_files}'+os.sep+'TN400_dgs_raw.Raw', 'Thompson', progressbar=False)
        self.assertEqual(
            dgs.iloc[0]['date_time'].timestamp(), 1647101046.634)
        self.assertEqual(dgs.iloc[0]['Gravity'], -82)
        self.assertTrue(dgs.iloc[0]['vcc'] - 0.0357100 < 0.0001)

    def test_read_dgs_raw_Sikuliaq(self):
        dgs = sgi.read_dgs_raw(f'{self.ex_files}'+os.sep+'SKQ_dgs_raw.20241018T2212Z', 'Sikuliaq', progressbar=False)
        self.assertEqual(
            dgs.iloc[0]['date_time'].timestamp(), 1729289555.6591)
        self.assertEqual(dgs.iloc[0]['rgrav'], 19964)
        self.assertTrue(dgs.iloc[0]['vcc'] - 0.30569 < 0.001)

    def test_read_mru(self):
        mru, cols = sgi.read_other_stuff(
            f'{self.ex_files}'+os.sep+'IXBlue.yaml', f'{self.ex_files}'+os.sep+'SR2312_mru.txt', 'PASHR')
        self.assertEqual(mru.iloc[0]['Pitch:g'], -0.41)
        self.assertEqual(mru.iloc[0]['Roll:g'], 2.03)
        self.assertEqual(mru.iloc[0]['Heave:g'], -0.6)


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(ioTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
Introduction
---------------------

shipgrav is a Python package with utilities for reading and processing marine gravity data from UNOLS ships. At time of writing, the UNOLS fleet is transitioning away from BGM3 gravimeters to DGS AT1M meters managed by the Potential Field Pool Equipment (PFPE) facility. shipgrav is able to read files from both types of meters, as well as navigation data and other vessel feeds. shipgrav functions can then be used to process gravity data to FAA, MBA, RMBA, and crustal thickness estimates.

DGS gravimeters output two types of files: serial, or 'raw' files; and 'laptop' files. Raw files are written from the serial port, and contain counts values that can be calibrated to retrieve the gravity signal. In this documentation we use the terms 'serial' and 'raw' interchangeably.  What we refer to as laptop files are lightly processed onboard the meter and output with (biased) gravity values alongside other information.

Installation
------------

shipgrav can be installed from `PyPI <https://pypi.org/project/shipgrav/>`_ using ``pip``. We recommend using an environment management tool like `conda <https://anaconda.org>`_. An exemplary set of commands to make a conda enviroment with shipgrav would be: ::

    conda create --name shipgrav numpy scipy pandas statsmodels tomli pyyaml tqdm pooch matplotlib geographiclib
    conda activate shipgrav
    pip install shipgrav

shipgrav's dependencies are

* Python 3.9+
* numpy
* scipy
* pandas 2.0+
* statsmodels
* tomli
* pyyaml
* tqdm
* pooch (optional, needed to run the example scripts)
* matplotlib (optional, needed to run some of the example scripts)
* geographiclib (optional, needed to run one of the example scripts)
* jupyterlab (optional, if you want to run example scripts in jupyter)
* jupytext (optional, if you want to run example scripts in jupyter)

If you install shipgrav with ``pip``, it will also install any of the required dependencies that are missing. To make ``pip`` include the optional dependencies, run ``pip install shipgrav[examples]``. Depending on your operating system and shell, you may need to escape the brackets (i.e. ``pip install shipgrav\[examples\]``).

The example scripts are `available on github <https://github.com/PFPE/shipgrav>`_. They are not packaged with the PyPI package and must be downloaded separately.

Modules and files
-----------------

shipgrav consists of the modules ``io``, ``nav``, ``grav``, and ``utils``, along with the file ``database.toml`` and a set of example scripts. 

* ``io`` contains functions for reading different kinds of gravimeter files and associated navigation files.
* ``nav`` contains functions for handling coordinate systems.
* ``grav`` contains functions for processing gravity data and calculating various anomalies.
* ``utils`` is a catch-all of other things we need. 
* ``database.toml`` holds some ship-specific constants and other information for UNOLS vessels.
* the scripts in ``example-scripts/`` walk through the steps of reading and processing UNOLS gravimeter data for a set of data files that are publicly available via `R2R (Rolling Deck to Repository) <https://www.rvdata.us>`_.

Data directories
----------------

You can organize your data however you like; shipgrav does not care as long as you tell it where to look. The example scripts are set up to download data files from public repositories using ``pooch``. The ``pooch.retrieve()`` function returns lists of file paths for files that have been downloaded and unpacked, so to adapt the example workflows for other data files, you will need to replace those lists of paths with the paths to your data.

Ties and bias
-------------

The ``database.toml`` file contains some bias values for UNOLS vessels. These are provided for your convenience, but are not necessarily up-to-date with recent gravity ties.

Navigation data
---------------

Which navigation data should you use to process gravimeter data?

In an ideal world, the gravimeter pulls navigation info from the ship's feed and synchronizes it perfectly with acquisition such that the output files have the correct geographic coordinates in them at the start. In practice, this synchronization doesn't always work as expected (see ``example-scripts/dgs_raw_comp.py`` for a case where the serial files do not have GPS info). So, we like to take the timestamped navigation data directly from the ship's feed and match up the gravimeter timestamps to obtain more accurate coordinates. Using navigation from a feed separate from the gravimeter is **highly recommended**.

The database file included in shipgrav lists the navigation talkers that we expect are good to use for specific UNOLS vessels. Find the files that contain those feeds, and you should be able to read in timestamped coordinates from them.

Example scripts
---------------

The scripts in the ``example-scripts/`` directory use publicly available data files to run through some common workflows for marine gravity processing. All of the examples can be run as scripts (ie, with ``python -m <script-name>.py``). All except ``interactive_line_pick.py`` can also be opened in ``jupyter`` as notebooks thanks to ``jupytext``. To run the examples in ``jupyter``, start ``jupyter lab``, right-click on the script file name, and select `open with -> notebook`.

The data files can be downloaded from R2R and Zenodo, and the scripts will do this automatically using ``pooch``. The sources are:

* https://doi.org/10.7284/151470 - TN400 BGM3 data
* https://doi.org/10.7284/151457 - TN400 nav data
* https://doi.org/10.7284/157179 - SR2312 DGS laptop data
* https://doi.org/10.7284/157188 - SR2312 nav data
* https://doi.org/10.7284/157177 - SR2312 mru data
* https://doi.org/10.5281/zenodo.12733929 - TN400 DGS raw and laptop data, SR2312 DGS raw data, R/V Ride DGS meter and Hydrins metadata, satellite FAA tracks for comparison, example file for RMBA calculation

``dgs_bgm_comp.py`` reads data from DGS and BGM gravimeter files from R/V Thompson cruise TN400. The files are lightly processed to obtain the FAA (including syncing with navigation data for more accurate locations), and the FAA is plotted alongside corresponding satellite-derived FAA.

.. image:: _static/TN400_FAA.png
   :alt: FAA for TN400 data from BGM3 and DGS, compared to satellite data.
   :height: 250px
   :align: center

``dgs_raw_comp.py`` reads laptop and serial data from R/V Sally Ride cruise SR2312. The serial data are calibrated and compared to the laptop data. The laptop data are processed to FAA and plotted alongside satellite-derived FAA.

.. image:: _static/SR2312_serial_laptop.png
   :alt: SR2312 raw gravity from serial and laptop files.
   :height: 250px
   :align: center

``dgs_ccp_calc.py`` reads DGS files from R/V Thompson cruise TN400, calculates the FAA and various kinematic variables, and fits for cross-coupling coefficients. The cross-coupling correction is applied and the data are plotted with and without correction. Satellite-derived FAA is also plotted

.. image:: _static/TN400_ccp.png
   :alt: FAA for TN400, with and without cross-coupling correction applied; and satellite-derived FAA.
   :height: 250px
   :align: center

``mru_coherence.py`` reads laptop data and other feeds from R/V Sally Ride cruise SR2312. The FAA is calculated, and MRU info is read to obtain time series of pitch, roll, and heave. Coherence is caluclated between those and each of the four monitors output by the gravimeter for the cross-coupling correction.

.. image:: _static/roll_coherence.png
   :alt: Coherence between monitors and roll for SR2312.
   :height: 250px
   :align: center

``interactive_line_pick.py`` reads laptop data and navigation data from R/V Sally Ride cruise SR2312. The script generates an interactive plot with a cursor for users to select segments of the time series data based on mapped locations, in order to extract straight line segments from a cruise track. `This script cannot be run in jupyter`. The selected segments are written to files that can be re-read by the next script...

.. image:: _static/cursor.png
   :alt: Interactive line segment picker window.
   :height: 250px
   :align: center

``RMBA_calc.py`` reads an example of data from a line segment (from the interactive line picker) and calculates the residual mantle bouger anomaly (RMBA) as well as estimated crustal thickness variations.

.. image:: _static/rmba.png
   :alt: RMBA for a segment of SR2312, with back-calculated "recovered" signal.
   :height: 250px
   :align: center

Help!
-----

``FileNotFound`` **errors:** check the filepaths in your scripts and make sure that (a) there are no typos, and (b) you are pointing toward the actual locations of your data files.

**Other file reading errors:** shipgrav does its best to read a variety of file formats from UNOLS gravimeters, but we can't read files that we don't know enough about ahead of time. In some cases, a file cannot be read because we don't yet know how to pass the file to the correct parsing function. Most primary i/o functions in shipgrav have an option where users can supply their own file-parsing function, so one option is to write such a function (following the examples in shipgrav for known vessel file formats) and plug that in via the appropriate kwarg (usually named ``ship_function``). You can also send an example file and information to PFPE so that we can update shipgrav.

**The anomaly I've calculated looks really weird:** a good first step is to compare your (lowpass filtered) FAA to satellite data (e.g., `Sandwell et al. 2014 <https://doi.org/10.1126/science.1258213>`_). If that looks very different, you can start checking whether the data is being read properly; whether the sample rate of the data is consistent with your expectations; whether there are anomalous spikes or dropouts in the data that need to be cleaned out; and whether the corrections used to calculate the FAA seem to have reasonable magnitudes.

**I want to use shipgrav, but my data is not from a UNOLS vessel:** the functions and workflows in shipgrav are entirely adaptable to use with data from other sources. You will need to determine the data format for your gravimeter files, and write or adapt a function to read that data. There are examples in the ``io`` module. If you have raw data files, you will also need to know the calibration constants and apply those. Once the data have been read (and calibrated), you should be able to apply all of the other shipgrav functions for processing. 

**I'm going to sea and want to be able to access this documentation offline:** this is all auto-generated from text included in the shipgrav source files! So one option is just to go read those (the main part of the documentation is in ``shipgrav/__init__.py``). To view it as a nice webpage, you can build the documentation locally using ``sphinx``. Install ``sphinx`` in your conda environment, run the command ``make html`` in the ``docs/`` directory, and then open ``docs/_build/html/index.html`` in your browser to view the documentation.

**If you have some other question that's not answered here:** you can try contacting PFPE at pfpe-internal(at)whoi.edu for specific assistance with processing UNOLS gravimeter data.

Testing
-------

shipgrav comes with a set of unit tests. To run them for yourself, navigate to the ``tests/`` directory and run ``__main__.py`` (in an environment with dependencies installed, naturally).


Contributing
------------

Do you have ideas for making this software better? Go ahead and `raise an issue <https://github.com/PFPE/shipgrav/issues>`_ on the github page or, if you're a savvy Python programmer, submit a pull request. You can also email PFPE at pfpe-interal(at)whoi.edu.

If you raise an issue on github, please include as much detail as possible, such as the text of error messages. If there are no visible errors but you think the code is behaving oddly, provide a description of what the code is doing and what you think it *should* be doing instead. PFPE may ask for additional details or copies of data files in order to reproduce and diagnose an issue.

"""



from copy import copy
from datetime import datetime, timezone
from math import factorial

import numpy as np
from pandas import DataFrame
from scipy.signal import firwin, lfilter
from scipy.special import erf, erfcinv
from statsmodels.api import OLS, add_constant

# impulse response of 10th order Taylor series differentiator
tay10 = [1/1260, -5/504, 5/84, -5/21, 5/6,
         0, -5/6, 5/21, -5/84, 5/504, -1/1260]

########################################################################
# tidal correction
########################################################################


def _convert_datetime_tidetime(timestamp):
    """Calculate julian century and hour from a timestamp.

    The reference point is noon on December 31, 1899 per Longman's paper.

    :param timestamp: time to convert to century + hour 
    :type timestamp: datetime.datetime

    :returns: 
        - **julian century** (*int*) - centuries since reference date, days/36525
        - **julian_hour** (*float*) - hours since midnight
    """
    origin_time = datetime(1899, 12, 31, 12, 0, 0, tzinfo=timezone.utc)
    dt = timestamp - origin_time
    days = dt.days + dt.seconds/3600./24.
    julian_century = days/36525
    julian_hour = (timestamp.hour + timestamp.minute /
                   60. + timestamp.second/3600.)

    return julian_century, julian_hour


def longman_tide_prediction(lon, lat, times, alt=0, return_components=False):
    """ Calculate predicted tidal effect on gravity using the Longman algorithm.

    The calculation is taken directly from

        Longman (1959). Formulas for Computing the Tidal Accelerations Due to
        the Moon and the Sun. Journal of Geophysical Research 64(12), 
        DOI: 10.1029/JZ064i012p02351

    as are all of the constant and variable descriptions.
    Tidal contribution(s) returned are in mGal.

    :param lon: longitude in decimal degrees, positive E
    :type lon: array_like
    :param lat: latitude in decimal degrees, positive N
    :type lat: array_like
    :param times: times for geographic locations
    :type times: array_like, datetime.datetime
    :param alt: altitude in meters (0 for sea level, for marine grav)
    :type alt: array_like
    :param return_components: if True, return lunar and solar components with total.
    :type return_components: bool

    :returns:
        - **g0** (*ndarray*)- total tidal effect in mGal
        - **gm** (*ndarray*)- lunar tidal effect in mGal (optional, if return_components is True)
        - **gs** (*ndarray*)- solar tidal effect in mGal (optional, if return_components is True)
    """

    assert len(lon) == len(lat), 'lengths of input vectors must be the same'
    assert len(lon) == len(times), 'lengths of input vectors must be the same'

    # convert the timestamps, referenced to Longman origin time
    T, t0 = np.empty(len(times)), np.empty(len(times))
    for i, stamp in enumerate(times):
        a, b = _convert_datetime_tidetime(stamp)
        T[i] = a
        t0[i] = b
        if t0[i] < 0:
            t0[i] += 24
        if t0[i] >= 24:
            t0[i] == 24

    TT = T*T    # squares and cubes are used a lot so shortcut them
    TTT = T*T*T

    # define a bunch of constants for the calculation
    mu = 6.673e-8  # Newton's gravitational constant
    M = 7.3537e25  # Mass of the moon in grams
    S = 1.993e33  # Mass of the sun in grams
    e = 0.05490  # Eccentricity of the moon's orbit
    m = 0.074804  # Ratio of mean motion of the sun to that of the moon
    c = 3.84402e10  # Mean distance between the centers of the earth and the moon
    c1 = 1.495e13  # Mean distance between centers of the earth and sun in cm
    h2 = 0.612  # Love parameter
    k2 = 0.303  # Love parameter
    a = 6.378270e8  # Earth's equitorial radius in cm
    i = 0.08979719  # (i) Inclination of the moon's orbit to the ecliptic
    # Inclination of the Earth's equator to the ecliptic 23.452 degrees
    omega = np.radians(23.452)
    lamb = np.radians(lat)  # (lambda) Latitude of point
    H = alt * 100.  # (H) Altitude above sea-level of point P in cm

    # Longman convention has W+/E- for unknown reasons
    L = -lon

    # Lunar gravity effects (Shureman 1941 coeffs)
    # mean longitude of the moon in its orbit, reckoned from the referred equinox
    s = 4.720023434 + 8399.709299*T + 0.0000440696*TT  # skip 3rd order bc it is tiny
    # mean longitude of lunar perigee
    p = 5.835124713 + 71.01800936*T - 0.000180545*TT - 0.00000021817*TTT
    # mean longitude of the sun
    h = 4.88162792 + 628.3319509*T + 0.00000527962*TT
    # longitude of moon's ascending node in orbit reckoned from referred equinox
    N = 4.523588564 - 33.75715303*T + 0.000036749*TT  # skip 3rd order bc it is tiny

    # inclination of the moon's orbit to the equator
    I = np.arccos(np.cos(omega)*np.cos(i) - np.sin(omega)*np.sin(i)*np.cos(N))
    # longitude in the celestial equator of its intersection A with the moon's orbit
    nu = np.arcsin(np.sin(i)*np.sin(N)/np.sin(I))
    # hour angle of mean sun measured westward from the place of observations
    t = np.radians(15. * (t0 - 12) - L)

    # right ascension of meridian of place of observations reckoned from A
    chi = t + h - nu
    # cos(alpha) where alpha is defined in eq. 15 and 16 of Longman 1959
    cos_alpha = np.cos(N)*np.cos(nu) + np.sin(N)*np.sin(nu)*np.cos(omega)
    # sin(alpha) where alpha is defined in eq. 15 and 16 of Longman 1959
    sin_alpha = np.sin(omega)*np.sin(N)/np.sin(I)
    # alpha from eq. 18 of Longman 1959
    alpha = 2*np.arctan(sin_alpha/(1 + cos_alpha))
    # longitude in the moon's orbit of its ascending intersection with the celestial equator
    xi = N - alpha

    # mean longitude of moon in radians in its orbit reckoned from A
    sigma = s - xi
    # longitude of moon in its orbit reckoned from its ascending intersection with the equator
    L_moon = (sigma + 2*e*np.sin(s - p) + (5./4)*e*e*np.sin(2*(s - p)) +
              (15./4)*m*e*np.sin(s - 2*h + p) + (11./8)*m*m*np.sin(2*(s - h)))

    # Solar calculations
    #  mean longitude of solar perigee
    p1 = 4.908229461 + 0.03000526416*T + 0.000007902463*TT  # skip tiny 3rd order term
    # eccentricity of Earth's orbit
    e1 = 0.01675104 - 0.0000418*T - 0.000000126*TT

    # right ascension of meridian of place of observations reckoned from the vernal equinox
    chi1 = t + h
    # longitude of sun in the ecliptic reckoned from the vernal equinox
    L_sun = h + 2*e1*np.sin(h - p1)
    # cosine(theta) where theta is the zenith angle of the moon
    cos_theta = (np.sin(lamb)*np.sin(I)*np.sin(L_moon) + np.cos(lamb)*(np.cos(0.5*I)**2
                                                                       * np.cos(L_moon - chi) + np.sin(0.5*I)**2*np.cos(L_moon + chi)))
    # cosine(phi) where phi is the zenith angle of the sun
    cos_phi = (np.sin(lamb)*np.sin(omega)*np.sin(L_sun) + np.cos(lamb) *
               (np.cos(0.5*omega)**2*np.cos(L_sun - chi1) +
               np.sin(0.5*omega)**2*np.cos(L_sun + chi1)))

    # Distance
    # distance parameter, eq. 34 in Longman 1959
    C = np.sqrt(1./(1 + 0.006738*np.sin(lamb)**2))
    # distance from point P to the center of the Earth
    r = C*a + H
    # distance parameter, eq. 31 in Longman 1959
    aprime = 1. / (c * (1 - e * e))
    # distance parameter, eq. 32 in Longman 1959
    aprime1 = 1. / (c1 * (1 - e1 * e1))
    # distance between centers of the Earth and the moon
    d = (1./((1./c) + aprime*e*np.cos(s - p) + aprime*e*e *
             np.cos(2*(s - p)) + (15./8)*aprime*m*e*np.cos(s - 2*h + p)
             + aprime*m*m*np.cos(2*(s - h))))
    # distance between centers of the Earth and the sun
    D = 1./((1./c1) + aprime1*e1*np.cos(h - p1))

    # vertical component of tidal acceleration due to the moon
    gm = ((mu*M*r/(d*d*d))*(3*cos_theta**2 - 1) + (3./2) *
          (mu*M*r*r/(d*d*d*d))*(5*cos_theta**3 - 3*cos_theta))
    # vertical component of tidal acceleration due to the sun
    gs = mu*S*r/(D*D*D)*(3*cos_phi**2 - 1)

    love = (1 + h2 - 1.5*k2)
    g0 = (gm + gs) * 1e3*love
    if return_components:
        return g0, gm*1e3*love, gs*1e3*love
    else:
        return g0

########################################################################
# Eotvos correction
########################################################################


def eotvos_full(lon, lat, ht, samp, a=6378137.0, b=6356752.3142):
    """ Full Eotvos correction in mGals.

    The Eotvos correction is the effect on measured gravity due to horizontal
    motion over the Earth's surface.

    This formulation is from Harlan (1968), "Eotvos Corrections for Airborne Gravimetry" in
    *Journal of Geophysical Research*, 73(14), DOI: 10.1029/JB073i014p04675

    Implementation modified from matlab script written by Sandra Preaux, NGS, NOAA August 24 2009

    Components of the correction:

    * rdoubledot
    * angular acceleration of the reference frame
    * corliolis
    * centrifugal
    * centrifugal acceleration of Earth

    :param lon: longitudes in degrees
    :type lon: array_like
    :param lat: latitudes in degrees
    :type lat: array_like
    :param ht: elevation (referenced to sea level)
    :type ht: array_like
    :param samp: sampling rate
    :type samp: float
    :param a: major axis of ellipsoid (default is WGS84)
    :type a: float, optional
    :param b: minor axis of ellipsoid (default is WGS84)
    :type b: float, optional

    :return: **E** (*ndarray*), Eotvos correction in mGal
    """
    We = 0.00007292115    # siderial rotation rate, radians/sec
    mps2mgal = 100000     # m/s/s to mgal
    ecc = (a-b)/a

    latr = np.deg2rad(lat)
    lonr = np.deg2rad(lon)

    # get time derivatives of position
    dlat = center_diff(latr, 1, samp)
    ddlat = center_diff(latr, 2, samp)
    dlon = center_diff(lonr, 1, samp)
    ddlon = center_diff(lonr, 2, samp)
    dht = center_diff(ht, 1, samp)
    ddht = center_diff(ht, 2, samp)

    # sines and cosines etc
    slat = np.sin(latr[1:-1])
    clat = np.cos(latr[1:-1])
    s2lat = np.sin(2*latr[1:-1])
    c2lat = np.cos(2*latr[1:-1])

    # calculate r' and its derivatives
    rp = a*(1 - ecc*slat*slat)
    drp = -a*dlat*ecc*s2lat
    ddrp = -a*ddlat*ecc*s2lat - 2*a*dlat*dlat*ecc*c2lat

    # calculate deviation from normal and derivatives
    D = np.arctan(ecc*s2lat)
    dD = 2*dlat*ecc*c2lat
    ddD = 2*ddlat*ecc*c2lat - 4*dlat*dlat*ecc*s2lat

    # define r and its derivatives
    r = np.vstack((-rp*np.sin(D), np.zeros(len(rp)), -
                  rp*np.cos(D) - ht[1:-1])).T
    rdot = np.vstack((-drp*np.sin(D) - rp*dD*np.cos(D),
                     np.zeros(len(rp)), -drp*np.cos(D) + rp*dD*np.sin(D) - dht)).T
    ci = -ddrp*np.sin(D) - 2.*drp*dD*np.cos(D) - rp * \
        (ddD*np.cos(D) - dD*dD*np.sin(D))
    ck = -ddrp*np.cos(D) + 2.*drp*dD*np.sin(D) + rp * \
        (ddD*np.sin(D) + dD*dD*np.cos(D)) - ddht
    rdotdot = np.vstack((ci, np.zeros(len(ci)), ck)).T

    # define w and derivative
    w = np.vstack(((dlon + We)*clat, -dlat, -(dlon + We)*slat)).T
    wdot = np.vstack((dlon*clat - (dlon + We)*dlat*slat, -
                     ddlat, -ddlon*slat - (dlon + We)*dlat*clat)).T

    w2xrdot = np.cross(2*w, rdot)
    wdotxr = np.cross(wdot, r)
    wxr = np.cross(w, r)
    wxwxr = np.cross(w, wxr)

    # calcualte wexwexre, centrifugal acceleration due to the Earth
    re = np.vstack((-rp*np.sin(D), np.zeros(len(rp)), -rp*np.cos(D))).T
    we = np.vstack((We*clat, np.zeros(len(slat)), -We*slat)).T
    wexre = np.cross(we, re)
    wexwexre = np.cross(we, wexre)
    wexr = np.cross(we, r)
    wexwexr = np.cross(we, wexr)

    # calculate total acceleration for the aircraft
    a = rdotdot + w2xrdot + wdotxr + wxwxr

    # Eotvos correction is the vertical component of the total acceleraton of
    # the aircraft minus the centrifugal acceleration of the Earth, convert to mGal
    E = (a[:, 2] - wexwexr[:, 2])*mps2mgal
    E = np.hstack((E[0], E, E[-1]))

    return E

########################################################################
# latitude correction functions
########################################################################


def free_air_second_order(lat, ht, free_water=False):
    """ 2nd order free-air correction

    :param lat: latitude, degrees
    :type lat: array_like
    :param height: elevation, meters
    :type height: array_like

    :return: free-air correction, mGal
    """
    s2lat = np.sin(np.deg2rad(lat))**2

    if not free_water:
        return -((0.3087691 - 0.0004398*s2lat)*ht) + 7.2125e-8*(ht**2)
    else:
        return -((0.22242 - 0.0004398*s2lat)*ht) + 7.2125e-8*(ht**2)


def wgs_grav(lat):
    """ Theoretical gravity for WGS84 ellipsoid

    :param lat: latitude, degrees
    :type lat: array_like

    :return: uniform ellipsoid gravity, mGal
    """
    sinsq = np.sin(np.deg2rad(lat))**2

    num = 1 + 0.00193185265241*sinsq   # something like (b*gp - a*ge)/(a*ge)
    den = np.sqrt(1 - 0.00669437999014*sinsq)  # this is e2
    return 978032.53359*(num/den)

########################################################################
# cross-coupling coefficients
########################################################################


def calc_cross_coupling_coefficients(faa_in, vcc_in, ve_in, al_in, ax_in, level_in, times=None, samplerate=1):
    """ Calculate cross-coupling coefficients from FAA via ordinary linear regression

    The cross-coupling coefficients are returned in `model`, in the `params` attribute

    :param faa_in: free air anomaly, filtered
    :type faa_in: array_like
    :param vcc_in: vcc monitor
    :type vcc_in: array_like
    :param ve_in: ve monitor
    :type ve_in: array_like
    :param al_in: al monitor
    :type al_in: array_like
    :param ax_in: ax monitor
    :type ax_in: array_like
    :param level_in: tilt/leveling correction, often negligible. 
        Use a vector of zeros to ignore this component.
    :type level_in: array_like
    :param times: timestamps to use for dividing the data into continuous sections
    :type times: array_like, floats, optional
    :param samplerate: used with times to detect large sampling gaps in the data
    :type samplerate: float, optional

    :return: 
        - **df** (*pd.DataFrame*)- double-differenced and filtered monitors and gravity
        - **model** (*statsmodels.OLS*)- linear regression model
    """

    end_inds = np.array([len(faa_in),])  # just the one
    if times is not None:  # supplied a vector of timestamps to go with everything else
        # split data into segments of continuous even sample rate
        tdiff = np.diff(times)
        if np.any(tdiff > 5*samplerate):
            # 5 sec gap doesn't matter much I hope???
            end_inds = np.where(tdiff > 5*samplerate)[0]

    end_inds = np.append(-1, end_inds)  # add starting point

    faa_tf = np.array([])
    vcc_tf = np.array([])
    ve_tf = np.array([])
    al_tf = np.array([])
    ax_tf = np.array([])
    le_tf = np.array([])  # things to fit

    for i in range(1, len(end_inds)):  # loop continuous-time segments
        faa = faa_in[end_inds[i-1]+1:end_inds[i]]
        vcc = vcc_in[end_inds[i-1]+1:end_inds[i]]
        ve = ve_in[end_inds[i-1]+1:end_inds[i]]
        al = al_in[end_inds[i-1]+1:end_inds[i]]
        ax = ax_in[end_inds[i-1]+1:end_inds[i]]
        level = level_in[end_inds[i-1]+1:end_inds[i]]
        if len(faa) < 1000:
            continue  # no point for very short segments
        # double derivatives with taylor series
        gpp = np.convolve(np.convolve(faa, tay10, 'same'), tay10, 'same')
        vccpp = np.convolve(np.convolve(vcc, tay10, 'same'), tay10, 'same')
        vepp = np.convolve(np.convolve(ve, tay10, 'same'), tay10, 'same')
        alpp = np.convolve(np.convolve(al, tay10, 'same'), tay10, 'same')
        axpp = np.convolve(np.convolve(ax, tay10, 'same'), tay10, 'same')
        levpp = np.convolve(np.convolve(level, tay10, 'same'), tay10, 'same')

        # trim the ends
        gpp = gpp[10:-10]
        vccpp = vccpp[10:-10]
        vepp = vepp[10:-10]
        alpp = alpp[10:-10]
        axpp = axpp[10:-10]
        levpp = levpp[10:-10]

        # fairly narrow filter to get rid of any high-freq noise generated by the double derivation
        filterlength = 100  # Aliod code has this as 10...but that gives VERY different cc values
        taps = int(2*filterlength)
        BM = firwin(taps, 1/taps, window='blackman')
        fgpp = lfilter(BM, 1, gpp)
        fvccpp = lfilter(BM, 1, vccpp)
        fvepp = lfilter(BM, 1, vepp)
        falpp = lfilter(BM, 1, alpp)
        faxpp = lfilter(BM, 1, axpp)
        flevpp = lfilter(BM, 1, levpp)

        # trim off filter transients
        fgpp = fgpp[taps:-taps]
        fvccpp = fvccpp[taps:-taps]
        fvepp = fvepp[taps:-taps]
        falpp = falpp[taps:-taps]
        faxpp = faxpp[taps:-taps]
        flevpp = flevpp[taps:-taps]

        faa_tf = np.append(faa_tf, fgpp)
        vcc_tf = np.append(vcc_tf, fvccpp)
        ve_tf = np.append(ve_tf, fvepp)
        al_tf = np.append(al_tf, falpp)
        ax_tf = np.append(ax_tf, faxpp)
        le_tf = np.append(le_tf, flevpp)

    faa_tf = np.array(faa_tf).flatten()
    vcc_tf = np.array(vcc_tf).flatten()
    ve_tf = np.array(ve_tf).flatten()
    al_tf = np.array(al_tf).flatten()
    ax_tf = np.array(ax_tf).flatten()
    le_tf = np.array(le_tf).flatten()

    # solve for curvature equation by simple regression (OLS)
    df = DataFrame({'ve': ve_tf, 'vcc': vcc_tf, 'al': al_tf,
                   'ax': ax_tf, 'lev': le_tf, 'g': -faa_tf})
    x = df[['ve', 'vcc', 'al', 'ax', 'lev']]
    y = df['g']
    x = add_constant(x)
    model = OLS(y, x).fit()
    return df, model

########################################################################
# things loosely connected to tilt corrections
########################################################################


def center_diff(y, n, samp):
    """ Numerical derivatives, central difference of nth order

    :param y: data to differentiate
    :type y: array_like
    :param n: order, either 1 or 2
    :type n: int
    :param samp: sampling rate
    :type samp: float

    :return: 1st or 2nd order derivative of **y**
    """
    if n == 1:
        return (y[2:] - y[:-2])*(samp/2)
    elif n == 2:
        return (y[:-2] - 2*y[1:-1] + y[2:])*(samp**2)
    else:
        print('bad order for derivative')
        return -999


def up_vecs(dt, g, cacc, lacc, on_off, cper, cdamp, lper, ldamp):
    """ Calculate 3xN matrix of platform up-pointing vectors in (cross, long) coordinates

    :param dt: sampling interval in seconds
    :type dt: float
    :param g: latitudinal correction term (2nd orfer FA plus ellipsoid)
    :type g: array_like
    :param cacc: cross-axis acceleration
    :type cacc: array_like
    :param lacc: long-axis acceleration
    :type lacc: array_like
    :param on_off: flag for "good" data - can be used to zero out data points when the meter
        is clamped or otherwise not operational
    :type on_off: array_like
    :param cper: platform period in seconds for the cross-axis tilt filter
    :type cper: float
    :param cdamp: platform damping term for cross-axis tilt filter
    :type cdamp: float
    :param lper: platform period in seconds for the long-axis tilt filter
    :type lper: float
    :param ldamp: platform damping term for long-axis tilt filter
    :type ldamp: float

    :return: **up_vecs** (*3xN ndarray*) - cross, long, and up vectors for the platform

    """
    # clean out any nans in the accelerations
    cacc[np.isnan(cacc)] = 0
    lacc[np.isnan(lacc)] = 0

    # apply the mysterious on_off
    cacc[on_off > 0] = 0
    lacc[on_off > 0] = 0

    # make the cross-axis tilt filter
    cnum, cden = _tilt_filter(cper, dt, cdamp)

    # calculate cross-axis driving term, do tilt filtering
    drive = cacc/g
    drive[np.isnan(drive)] = 0
    ctilt = lfilter(cnum, cden, drive)

    # repeat all that for the long axis
    lnum, lden = _tilt_filter(lper, dt, ldamp)
    drive = lacc/g
    drive[np.isnan(drive)] = 0
    ltilt = lfilter(lnum, lden, drive)

    # combine the pieces to get the platform up-pointing vectors
    up_vecs = _calc_up_vecs(np.arctan(ctilt), np.arctan(ltilt))

    return up_vecs


def _tilt_filter(per, dt, damp=False):
    """ Filter coefficients for L&R platform tilt computation

    :param per: platform period, seconds
    :type per: float
    :param dt: sample increment, seconds
    :type dt: float
    :param damp: platform damping term (default: sqrt(2)/2)
    :type damp: float, optional

    :return: filter coefficients
    """

    if not damp:
        damp = np.sqrt(2)/2

    # frequencies:
    w0 = (2*np.pi)/per
    ws = 1/np.sqrt(6371100/9.8)

    om0 = (2/dt)*np.tan(w0*dt/2)
    omS = (2/dt)*np.tan(ws*dt/2)

    # first stage substitutions
    a = (omS**2) - (om0**2)
    b = 2*damp*om0*(2/dt)
    c = 4/(dt**2)
    d = om0**2

    # second stage substitutions
    d0 = b + c + d
    b0 = (a - b)/d0
    b1 = 2*a/d0
    b2 = (a + b)/d0
    a1 = 2*(d - c)/d0
    a2 = (c + d - b)/d0

    num = np.array([b0, b1, b2])
    den = np.array([1, a1, a2])

    return num, den


def _calc_up_vecs(ctilt, ltilt):
    """ calculate 3xN matrix of platform up-vectors in (cross, long, up) coordinates

    :param ctilt: cross-axis tilt angles in radians
    :type ctilt: array_like
    :param ltilt: long-axis tilt angles in radians
    :type ltilt: array_like

    :return: **up_vecs** (*3xN ndarray*) - (cross, long, up) for platform
    """

    # trig functions of tilts
    sc = np.sin(ctilt)
    cc = np.cos(ctilt)
    sl = np.sin(ltilt)
    cl = np.cos(ltilt)

    # set up array to hold output
    n = len(ctilt)
    up_vecs = np.zeros((3, n))

    # do rotations
    for i in range(n):
        # rotation matrics
        rotc = np.array([[cc[i], 0, -sc[i]], [0, 1, 0], [sc[i], 0, cc[i]]])
        rotl = np.array([[1, 0, 0], [0, cl[i], -sl[i]], [0, sl[i], cl[i]]])

        # alternate order of rotations for reasons that I do not understand
        plat_up = [0, 0, 1]  # start with vertical platform
        if i % 2 == 0:
            plat_up = np.matmul(rotc, np.matmul(rotl, plat_up))
        else:
            plat_up = np.matmul(rotl, np.matmul(rotc, plat_up))

        up_vecs[:, i] = plat_up

    return up_vecs

########################################################################
# MBA and RMBA functions (including thermal models)
########################################################################


def grav1d_padded(xtopo, topo, zlev, rho):
    """Calculate the gravity anomaly due to a density contrast across topography, along a line.

    The input (1D) topography is padded on both ends to reduce edge effects

    This function uses the method from Parker and Blakely:

        R. L. Parker (1972). The Rapid Calculation of Potential Anomalies,
        Geophys J R astr Soc 31, 447-455, DOI: 10.1111/j.1365-246X.1973.tb06513.x

        R. J. Blakely (1995). "Ch. 11: Fourier-Domain Modeling" in **Potential Theory in Gravity 
        and Magnetic Applications**, Cambridge University Press, DOI: 10.1017/CBO9780511549816

    .. Original 2D function written by Mark Behn, November 6, 2003
       Translated to Python in 1D with padding by Hannah Mark, 6 October 2017

    :param xtopo: x coordinates of the surface in meters (must be equally spaced)
    :type xtopo: array_like
    :param topo: z coordinates of the surface in meters
    :type ztopo: array_like
    :param zlev: vertical distance for upwards continuation in meters.
    :type zlev: float
    :param rho: density contrast across the topography in kg/m^3.
    :type rho: float

    :return: **anom** (*ndarray*) - gravity anomaly in mgal
    """
    G = 6.673*1e-8
    grav = 2*np.pi*G*rho
    baselev = np.mean(topo)  # mean topography for baseline

    # spacing of coordinates, must be constant
    dx = (xtopo[-1]-xtopo[0])/(len(xtopo)-1)
    nx = len(xtopo)

    wing = np.ones(2*len(topo))  # *** extra padding

    # extend or mirror the profile
    padtopo = np.append(topo[0]*wing, np.append(topo, topo[-1]*wing))
    # padtopo = np.append(baselev*wing,np.append(topo,baselev*wing))
    # padtopo = np.append(topo,topo[::-1])
    nxt = len(padtopo)
    mfx = (nxt/2.) + 1

    k = (2*np.pi)/(nxt*dx)  # calculate wavenumbers
    k2 = k**2
    xi1 = np.arange(1, mfx+1)
    xi = np.append(xi1, xi1[-2:0:-1])
    xxk = (xi-1)*(xi-1)*k2
    kwn1 = np.sqrt(xxk[:nxt])

    Ftopo = np.fft.fft(padtopo)  # Fourier transform the topography
    Ftopo[0] = 0

    npower = 5
    SUMtopo = copy(Ftopo)
    for ip in range(2, npower+1):  # summation per eq. 11.41 in Blakely
        Ftopo = np.fft.fft(padtopo**ip)
        Ftopo = Ftopo*(kwn1**(ip-1))/factorial(ip)

        SUMtopo = SUMtopo + Ftopo

    data = SUMtopo*grav*np.exp(-(zlev-baselev)*kwn1)  # upward continuation

    data = np.fft.ifft(data)  # inverse Fourier transform
    anom = np.real(data[2*nx:3*nx])*100  # factor of 100 for mgal output

    return anom


def grav2d_folding(X, Y, Z, dx, dy, drho=0.6, dz=6000, ifold=True, npower=5):
    """
    Parker [1972] method for calculating gravity from 2D topographic surface with a density contrast.

    The `ifold` option enables folding the input topography grid in x and y to mitigate edge effects

    .. Modified from parker.m by Mark Behn, which was in turn modified from
       parker.f by Ban-Yuan Kuo and Jian Lin.

    :param X: vector of N X coordinates
    :type X: ndarray
    :param Y: vector of M Y coordinates
    :type Y: ndarray
    :param Z: matrix of Z coordinates
    :type Z: ndarray, NxM
    :param dx: x grid spacing, for wavenumbers [km]
    :type dx: float
    :param dy: y grid spacing, for wavenumbers [km]
    :type dy: float
    :param drho: density contrast across surface. Ex: 1.7 for water to crust, 0.6 for crust to mantle
    :type drho: float
    :param dz: offset depth for layer interface, added to baselevel for upward continuation [m]
    :type dz: float
    :param ifold: switch for folding
    :type ifold: bool
    :param npower: power of Taylor series expansion (default: 5)
    :type npower: int

    :return: (*ndarray*) gravity anomaly in mgal
    """
    nx = len(X)  # size of the grid
    ny = len(Y)

    zmax = max(Z.flatten())   # get min and max for scaling
    zmin = min(Z.flatten())
    if zmax < -1e10:
        zmax = -1e10
    if zmin > 1e10:
        zmin = 1e10

    slev = np.mean(Z.flatten())

    Z = 100*(slev-Z)
    slev = slev*100
    dx = 100000*dx
    dy = 100000*dy
    dz = dz*100

    # to fold or not to fold--
    if ifold:
        nxt = nx*2
        nyt = ny*2
    else:
        nxt = nx
        nyt = ny

    G = 6.673e-8   # gravitational constant
    conv = 1000  # gals to mgals
    grav1 = 2*np.pi*G*conv

    grav = grav1*drho

    # for wavenumbers
    kint1 = (2*np.pi)/(nxt*dx)
    kint2 = (2*np.pi)/(nyt*dy)
    kx2 = kint1*kint1
    ky2 = kint2*kint2

    # folding frequency:
    mfx = int(nxt/2 + 1)
    mfy = int(nyt/2 + 1)

    # the important part:
    # (1) compute wavenumbers
    # (2) transform topography with fft
    # (3) sum over powers with those wavenumbers
    # (4) upward continue, transform back to space domain, multiply
    #     by constants

    # wavenumbers:
    yj1 = np.arange(1, mfy+1)
    yj = np.append(yj1, yj1[mfy-2:0:-1])
    yyk = (yj-1)*(yj-1)*ky2

    xi1 = np.arange(1, mfx+1)
    xi = np.append(xi1, xi1[mfx-2:0:-1])
    xxk = (xi-1)*(xi-1)*kx2

    kwn1 = np.zeros((nxt, nyt))
    for j in range(nyt):
        kwn1[:, j] = np.sqrt(xxk + yyk[j])

    kwn1 = kwn1.T

    # fold the data
    if ifold == 1:
        Z = np.vstack((Z, np.flipud(Z)))  # fold in X
        Z = np.hstack((Z, np.fliplr(Z)))  # fold in Y

    # first power
    data = np.copy(Z)
    data = np.fft.fft2(data)
    data[0, 0] = 0

    # sum over the other powers
    csum = np.copy(data)  # for adding summation terms
    fact = 1
    for i in range(2, npower+1):
        fact = fact*i
        data = np.copy(Z)**i
        data = np.fft.fft2(data)

        data = data*(kwn1**(i-1))/fact
        k, l = np.where(kwn1 == 0)
        data[k, l] = 0

        csum = csum + np.copy(data)

    # upward continuation
    zlev = slev+dz
    data = csum*grav*np.exp(-zlev*kwn1)

    data = np.fft.ifft2(data)
    sdata = np.real(data[:ny, :nx])  # back in the spatial domain

    return sdata


def grav2d_layer_variable_density(rho, dx, dy, z1, z2):
    """
    Calculate the gravity contribution from a layer of equal thickness with
    an inhomogenous density distribution in x and y (homogeneous in z)

    .. Based on glayer.m by Mark Behn

    :param rho: 2D density distribution [kg/m^3]
    :type rho: ndarray
    :param dx,dy: sample intervals in km
    :type dx,dy: float
    :param z1,z2: depth to top and bottom of layer in km (both >0)
    :type z1,z2: float

    :return: (*ndarray*) gravity in mgal
    """

    si2mg = 1e5
    km2m = 1e3
    G = 6.673e-11
    grav = 2*np.pi*G

    ny, nx = rho.shape
    dkx = 2*np.pi/(nx*dx)
    dky = 2*np.pi/(ny*dy)

    ifrho = np.fft.fft2(rho)  # take 2D fft of densities

    crho = np.empty((ny, nx), dtype=complex)
    for j in range(nx):
        for i in range(ny):
            kx, ky = _kvalue(i, j, nx, ny, dkx, dky)
            k = np.sqrt(kx**2 + ky**2)
            if k == 0:
                crho[i, j] = 0
            else:
                crho[i, j] = ifrho[i, j]*grav*(np.exp(-k*z1)-np.exp(-k*z2))/k

    grho = np.fft.ifft2(crho)
    grho = np.real(grho)*si2mg*km2m

    return grho


def _kvalue(i, j, nx, ny, dkx, dky):
    """
    Get wavenumber coordinates of one element of a rectangular grid

    :param i,j: indices in ky,kx directions
    :type i,j: int
    :param nx,ny: dimensions of the grid in the ky,kx directions
    :type nx,ny: int
    :param dkx,dky: sample intervals in kx,ky directions
    :type dkx,dky: float

    :return: **kx, ky** (*float*) - x and y wavenumbers at (i, j)
    """

    nyqx = nx/2+1
    nyqy = ny/2+1

    if j <= nyqx:
        kx = (j)*dkx
    else:
        kx = (j-nx)*dkx

    if i <= nyqy:
        ky = (i)*dky
    else:
        ky = (i-ny)*dky
    return kx, ky


def therm_halfspace(x, z, u=0.01, Tm=1350, time=False, rhom=3300, rhow=1000,
                    a=3.e-5, k=1.e-6):
    """Calculate thermal structure for a half space cooling model.

    Reference:

        D. Turcotte & G. Schubert (2014). Geodynamics. Cambridge
        University Press. DOI: 10.1017/CBO9780511843877
        Relevant pages: 161-162, 174-176 in 2nd or 3rd ed?

    .. Written by Mark D. Behn, November 20, 2003.
       Translated to Python + modded for plate age by Hannah Mark, October 2017

    :param x: vector of across-axis distance (meters) OR of plate ages (Myr)
    :type x: array_like
    :param z: vector of depth (meters)
    :type z: array_like
    :param u: spreading rate (m/yr)
    :type u: float
    :param time: switch for x vs age input: if ages, set time=True
        and u will be ignored.
    :type time: bool
    :param rhom: mantle density, kg/m^3, default 3300
    :type rhom: float, optional
    :param rhow: water density, kg/m^3, default 1000
    :type rhow: float, optional
    :param a: coefficient of thermal expansion, m^2/sec, default 3e-5
    :type a: float, optional
    :param k: thermal diffusivity, m^2/sec, default 1e-6
    :type k: float, optional

    :return:
        - **T** (*ndarray*) - gridded temperature over (x, z)
        - **W** (*ndarray*) - seafloor subsidence in meters
    """

    To = 0  # surface temperature [K]
    # a = 6.e-5  # coeff of thermal expansion [m^2/sec]  # used for SCARF calcs (???)
    # k = 2.e-6  # thermal diffusivity [m^2/sec]

    secyr = 365.25*24*3600  # seconds per year

    if time:
        x = x*1e6*secyr  # convert Myr to sec
    elif not time:
        x = abs(x)/(u/secyr)  # convert m to sec

    X, Z = np.meshgrid(x, z)  # grid up x,z pairs

    # mantle temperature
    T = To + (Tm-To) * erf(Z/(2*(k*X)**.5))

    # seafloor subsidence
    W = ((2*rhom*a*(Tm-To))/(rhom-rhow)) * \
        (((k*X[0, :])/(np.pi))**.5)

    return T, W


def therm_Z_halfspace(x, T, u=0.01, Tm=1350, time=False, rhom=3300, rhow=1000,
                      a=3.e-5, k=1.e-6):
    """Calculate depth to an isotherm for a half-space cooling model.

    :param x: vector of across-axis distance [m] OR plate age [Myr]
    :type x: array_like
    :param T: isotherm of choice [K]
    :type T: float
    :param u: spreading rate [m/yr], default 0.01
    :type u: float
    :param time: switch for x vs age input - if time=True,
        u is ignored, default False
    :type time: bool
    :param Tm: mantle potential temperature [K], default 1350
    :type Tm: float, optional
    :param rhom: mantle density, kg/m^3, default 3300
    :type rhom: float, optional
    :param rhow: water density, kg/m^3, default 1000
    :type rhow: float, optional
    :param a: coefficient of thermal expansion, m^2/sec, default 3e-5
    :type a: float, optional
    :param k: thermal diffusivity, m^2/sec, default 1e-6
    :type k: float, optional

    :returns:
        - **Z** (*ndarray*) - depth of this isotherm below the seafloor in meters
        - **W** (*ndarray*) - seafloor subsidence in meters
    """

    To = 0  # surface temperature [K]

    secyr = 365.25*24*3600  # seconds per year

    if time:
        x = x*1e6*secyr  # convert Myr to sec
    elif not time:
        x = abs(x)/(u/secyr)  # convert m to sec

    Z = 2*np.sqrt((k*x))*erfcinv((T-Tm)/(To-Tm))

    W = ((2*rhom*a*(Tm-To))/(rhom-rhow)) * \
        (((k*x)/(np.pi))**.5)

    return Z, W


def therm_plate(x, z, u=0.01, zL0=100.e3, Tm=1350, time=False, rhom=3300, rhow=1000,
                a=3.e-5, k=1.e-6):
    """Calculate thermal structure for the plate cooling model.  

    Reference:

        D. Turcotte & G. Schubert (2014). Geodynamics. Cambridge
        University Press. DOI: 10.1017/CBO9780511843877

    .. Written by Mark D. Behn, November 20, 2003.
       Translated to Python + modded for plate age by Hannah Mark, October 2017

    :param x: vector of across-axis distance (meters) OR plate age (Myr)
    :type x: array_like
    :param z: vector of depth (meters)
    :type z: array_like
    :param u: spreading rate (m/yr), default 0.01
    :type u: float
    :param zL0: plate thickness (meters), default 100e3
    :type zL0: float
    :param time: switch for x vs age input. If time=True, u is ignored.
    :type time: bool
    :param Tm: mantle potential temperature (K), default 1350
    :type Tm: float, optional
    :param rhom: mantle density, kg/m^3, default 3300
    :type rhom: float, optional
    :param rhow: water density, kg/m^3, default 1000
    :type rhow: float, optional
    :param a: coefficient of thermal expansion, m^2/sec, default 3e-5
    :type a: float, optional
    :param k: thermal diffusivity, m^2/sec, default 1e-6
    :type k: float, optional

    :returns:
        - **T** (*ndarray*) - gridded temperature over (x, z)
        - **W** (*ndarray*) - seafloor subsidence in meters
    """

    To = 0  # surface temperature [K]

    secyr = 365.25*24*3600  # seconds per year

    X, Z = np.meshgrid(x, z)

    if time:
        t = X*1e6*secyr  # convert Myr to sec
    elif not time:
        t = abs(X)/(u/secyr)  # convert across-axis distance to time in SECONDS

    Tterm2 = (2/np.pi)*np.exp(-k*(np.pi**2)*t/(zL0**2))*np.sin(np.pi*Z/zL0)
    Tterm3 = (1/np.pi)*np.exp(-4*k*(np.pi**2)*t/(zL0**2))*np.sin(2*np.pi*Z/zL0)
    Tterm4 = (2/np.pi/3)*np.exp(-9*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(3*np.pi*Z/zL0)
    Tterm5 = (2/np.pi/4)*np.exp(-16*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(4*np.pi*Z/zL0)
    Tterm6 = (2/np.pi/5)*np.exp(-25*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(5*np.pi*Z/zL0)
    Tterm7 = (2/np.pi/6)*np.exp(-36*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(6*np.pi*Z/zL0)
    Tterm8 = (2/np.pi/7)*np.exp(-49*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(7*np.pi*Z/zL0)
    Tterm9 = (2/np.pi/8)*np.exp(-64*k*(np.pi**2)
                                * t/(zL0**2))*np.sin(8*np.pi*Z/zL0)
    Tterm10 = (2/np.pi/9)*np.exp(-81*k*(np.pi**2)
                                 * t/(zL0**2))*np.sin(9*np.pi*Z/zL0)
    Tterm11 = (2/np.pi/10)*np.exp(-100*k*(np.pi**2)
                                  * t/(zL0**2))*np.sin(10*np.pi*Z/zL0)

    T = To + (Tm - To)*(Z/zL0 + Tterm2 + Tterm3 + Tterm4 + Tterm5 + Tterm6 + Tterm7 +
                        Tterm8 + Tterm9 + Tterm10 + Tterm11)

    mantle = np.where(Z > zL0)[0]
    T[mantle] = Tm

    Wterm2 = (4/np.pi**2)*np.exp(-k*(np.pi**2)*t[0, :]/(zL0**2))
    Wterm3 = (4/(9*np.pi**2))*np.exp(-k*9*(np.pi**2)*t[0, :]/(zL0**2))
    Wterm4 = (4/(25*np.pi**2))*np.exp(-k*25*(np.pi**2)*t[0, :]/(zL0**2))

    W = ((rhom*a*(Tm-To)*zL0)/(rhom-rhow)) * (1./2 - Wterm2 - Wterm3 - Wterm4)

    return T, W


def therm_Z_plate(x, T, u=0.01, zL0=100.e3, Tm=1350, time=False,
                  minz=0, maxz=100e3, zsp=1e2, rhom=3300, rhow=1000,
                  a=3.e-5, k=1.e-6):
    """Calculate approximate depth to an isotherm in the plate cooling model

    This is done by calculating a temperature field with a decent z spacing 
    and finding the closest points to the isotherm, so it depends strongly 
    on the z spacing that you use. If you need depth to multiple isotherms
    it's most efficient to get them all at once (using a longer array for
    T) so the whole temperature field is only calculated one time.

    :param x: array of across-axis distance (meters) OR plate age (Myr)
    :type x: array_like
    :param T: temperatures for which you want isotherms (K)
    :type T: array_like
    :param u: spreading rate (m/yr), default 0.01
    :type u: float
    :param zL0: plate thickness (meters), default 100e3
    :type zL0: float
    :param Tm: mantle potential temperature (K), default 1350
    :type Tm: float, optional
    :param time: switch for x vs age input. If time=True, u is ignored.
    :type time: bool
    :param minz: minimum z for calculating T field (meters), default 0
    :type minz: float
    :param maxz: maximum z (meters), default 100e3
    :type maxz: float
    :param zps: z spacing for grid (meters), default 1e3
    :type zps: float
    :param rhom:  mantle density, kg/m^3, default 3300
    :type rhom: float, optional
    :param rhow: water density, kg/m^3, default 1000
    :type rhow: float, optional
    :param a: coefficient of thermal expansion, m^2/sec, default 3e-5
    :type a: float, optional
    :param k: thermal diffusivity, m^2/sec, default 1e-6
    :type k: float, optional

    :returns: **ziso** (*ndarray*) - depths to isotherms, z(T,x)
    """

    z = np.arange(minz, maxz, zsp)

    Tp, _ = therm_plate(x, z, u=u, zL0=zL0, Tm=Tm, time=time, rhom=rhom, rhow=rhow,
                        a=a, k=k)  # calculate the whole temperature field

    ziso = np.zeros((len(T), len(x)))
    for i in range(len(T)):  # loop isotherms, find depths
        for j in range(len(x)):
            ziso[i, j] = z[np.argmin(abs(Tp[:, j]-T[i]))]

    return ziso

########################################################################
# crustal thickness functions
########################################################################


def crustal_thickness_2D(ur, nx=1000, ny=1, dx=1.3, dy=0, zdown=10, rho=0.4,
                         wlarge=45, wsmall=25, back=False):
    """
    Downward continuation of gravity to "topographic relief" ie crustal thickness

    This can be used in 2D, but also works for a single line
    given ny=1 (which is the default)

    .. Written by Hannah Mark (MIT/WHOI), October 2017
       Modeled on down_2d.f by Jian Lin (WHOI)

    :param ur: residual gravity anomaly, mgal
    :type ur: array_like
    :param nx: number of points in x direction, default 1000
    :type nx: int
    :param ny: number of points in y direction, default 1 (>1 for 2D)
    :type ny: int
    :param dx: spacing between x points, km, default 1.3
    :type dx: float
    :param dy: spacing between y points, km, default 0 (>0 for 2D)
    :type dy: float
    :param zdown: downward continuation depth, km, default 10
    :type zdown: float
    :param rho: density difference crust to mantle, g/cm^3, default 0.4
    :type rho: float
    :param wlarge: max wavelength for taper/cutoff, km, default 45
    :type wlarge: float
    :param wsmall: min wavelength for taper/cutoff, km, default 25
    :type wsmall: float
    :param back: switch for doing reverse tranform
    :type back: bool

    :returns:
        - **crustal thickness** (*ndarray*) - thickness variation in km
        - **recovered gravity** (*ndarray*) - back-calculated RMBA, (optional, if back=True)
    """
    assert wlarge > wsmall, 'wlarge must be larger than wsmall'

    ave = np.mean(ur)

    # shift to a new reference and convert milligal to gal
    # sign is switched so that positive residual = crustal thinning
    ur = -0.001*(ur - ave)

    # convert km to cm for spatial params
    dx = dx*1e5
    dy = dy*1e5
    zdown = zdown*1e5

    # calculate wavenumbers for taper/cutoff
    # we will taper btwn kcut1 and kcut2; cutoff >kcut2
    wlarge = wlarge*1e5   # convert km to cm
    wsmall = wsmall*1e5
    kcut1 = 2*np.pi/wlarge
    kcut2 = 2*np.pi/wsmall
    dkcut = kcut2 - kcut1
    mfx = nx/2 + 1
    mfy = ny/2 + 1

    prod = 1./(nx*ny)  # dimension correction factor

    G = 6.673e-8  # cgs gravity
    topo1 = 1/(2*np.pi*G*rho)  # gravity-to-topography transfer

    kint1 = 2*np.pi/(nx*dx)
    if dy != 0:
        kint2 = 2*np.pi/(ny*dy)
    else:
        kint2 = 0

    kwn1 = np.zeros((nx, ny))  # compute wavenumbers in 2D
    for j in range(ny):
        yj = j
        if j > mfy:
            yj = mfy*2 - j
        yyk = kint2**2 * (yj-1)**2

        for i in range(nx):
            xi = i
            if i > mfx:
                xi = mfx*2 - i
            xxk = kint1**2 * (xi-1)**2

            kwn1[i, j] = np.sqrt(xxk + yyk)

    # Fourier transform of the gravity residual
    # this was passed as a 1D array even for a 2D problem
    ur_arr = ur.reshape(nx, ny)
    ur_arr_ft = np.fft.fft2(ur_arr)  # 2D fft

    # apply wavenumbers, taper
    for j in range(ny):
        for i in range(nx):
            wgt = 1
            if kwn1[i, j] > kcut2:
                wgt = 0
            elif kwn1[i, j] > kcut1 and kwn1[i, j] <= kcut2:
                t = (kwn1[i, j]-kcut1)/dkcut*np.pi
                wgt = (np.cos(t)+1)/2

            ur_arr_ft[i, j] = ur_arr_ft[i, j] * \
                np.exp(kwn1[i, j]*zdown)*topo1*wgt

    # inverse Fourier transform
    ur_arr_2 = np.fft.fft2(ur_arr_ft)

    # back to vector, correct for dimensions, convert to km
    ur_arr_2 = ur_arr_2.reshape(-1, 1)
    ur_arr_2 = ur_arr_2*prod/1e5

    if not back:
        return ur_arr_2
    elif back:  # reverse the transform and upward continute to check gravity recovery
        for j in range(ny):
            for i in range(nx):
                ur_arr_ft[i, j] = -ur_arr_ft[i, j] * \
                    np.exp(-kwn1[i, j]*zdown)/topo1
        ur_back = np.fft.fft2(ur_arr_ft)
        ur_back = ur_back.reshape(-1, 1)
        ur_back = ur_back*prod*1000+ave

        return ur_arr_2[::-1], ur_back[::-1]



"""
Tests for shipgrav.nav
"""
import unittest

import numpy as np
import shipgrav.nav as sgn


class navTestCase(unittest.TestCase):
    def setUp(self):
        lons = np.ones(100)*70
        lats = np.linspace(40, 41, 100)
        # lat/lon to easting and northing *velocities*
        ve, vn = sgn.latlon_to_EN(lons, lats)
        self.ve = ve
        self.vn = vn

    def test_ll2en(self):
        self.assertTrue(self.ve[5] - 1121.5717 < 0.001)
        self.assertTrue(self.vn[5] < 1e-7)

    def test_en2cv(self):
        course, vel = sgn.ENvel_to_course_heading(self.ve, self.vn)
        self.assertTrue(course[5] - 90. < 0.001)
        self.assertEqual(vel[5], self.ve[5])

    def test_rot_acc(self):
        course, vel = sgn.ENvel_to_course_heading(self.ve, self.vn)
        eacc = 1e5*np.convolve(self.ve, sgn.tay10, 'same')
        nacc = 1e5*np.convolve(self.vn, sgn.tay10, 'same')
        cross, long = sgn.rotate_acceleration_EN_to_cl(course, eacc, nacc)
        self.assertTrue(cross[10] < 1e-7)
        self.assertEqual(long[10], eacc[10])


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(navTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
Tests for shipgrav.grav
These use 1000 seconds of sample data to calculate corrections and check values
ccp coeffs are not objectively good because the time series is short but it works for tests
Not bothering with the leveling correction here because exactly how to calculate it
is left to the user
"""
import os
import unittest
from glob import glob

import numpy as np
import shipgrav.grav as sgg
import shipgrav.io as sgi
import shipgrav.nav as sgn


class gravDataTestCase(unittest.TestCase):
    @property
    def ex_files(self):
        return __file__.rsplit(os.sep, 1)[0] + os.sep + "ex_files"

    def setUp(self):  # actions to take before running each test: load in some test data
        # test data is part of one of the files supplied by DGS (original file is 30M/86400
        # lines for 24 hr; we cut to 1 hr?)
        gfiles = glob(f'{self.ex_files}'+os.sep+'DGStest*.dat')
        data = sgi.read_dgs_laptop(gfiles, 'DGStest')
        data['tsec'] = [e.timestamp()
                        for e in data['date_time']]  # get posix timestamps
        data['grav'] = data['rgrav'] + 969143
        self.data = data

    def test_longman(self):
        lt = sgg.longman_tide_prediction(
            self.data['lon'], self.data['lat'], self.data['date_time'])
        self.assertTrue(lt[0] + 0.079599 < 0.001)

    def test_eotvos(self):
        eotvos = sgg.eotvos_full(self.data['lon'].values, self.data['lat'].values,
                                 np.zeros(len(self.data)), 1)
        self.assertTrue(eotvos[0] + 56.90367 < 0.001)

    def test_fa2ord(self):
        fa2 = sgg.free_air_second_order(
            self.data['lat'], np.zeros(len(self.data)))
        self.assertEqual(fa2.iloc[0], 0.)

    def test_up_vecs(self):
        lat_corr = sgg.wgs_grav(self.data['lat']) + sgg.free_air_second_order(self.data['lat'],
                                                                              np.zeros(len(self.data)))
        ve, vn = sgn.latlon_to_EN(
            self.data['lon'].values, self.data['lat'].values)
        eacc = 1e5*np.convolve(ve, sgn.tay10, 'same')
        nacc = 1e5*np.convolve(vn, sgn.tay10, 'same')
        crse, vel = sgn.ENvel_to_course_heading(ve, vn)
        acrss, along = sgn.rotate_acceleration_EN_to_cl(crse, eacc, nacc)
        up_vecs = sgg.up_vecs(1, lat_corr, acrss, along,
                              0, 240, 0.7071, 240, 0.7071)
        self.assertEqual(up_vecs[0, 0], 0.)
        self.assertEqual(up_vecs[2, 0], 1.)

    def test_ccpcalc(self):
        lt = sgg.longman_tide_prediction(
            self.data['lon'], self.data['lat'], self.data['date_time'])
        eotvos = sgg.eotvos_full(self.data['lon'].values, self.data['lat'].values,
                                 np.zeros(len(self.data)), 1)
        lat_corr = sgg.wgs_grav(self.data['lat']) + sgg.free_air_second_order(self.data['lat'],
                                                                              np.zeros(len(self.data)))
        faa = self.data['grav'] - lat_corr + eotvos + lt
        _, model = sgg.calc_cross_coupling_coefficients(faa, self.data['vcc'].values, self.data['ve'].values,
                                                        self.data['al'].values, self.data['ax'].values, np.zeros(len(self.data)))
        self.assertTrue(model.params.ax + 3.428237 < 0.0001)
        self.assertTrue(model.params.ve - 0.055732 < 0.0001)


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(gravDataTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
Tests for shipgrav
"""

import importlib.resources as importlib_resources
import sys
import unittest


def run():
    loader = unittest.TestLoader()
    ref = importlib_resources.files('shipgrav') / 'tests'
    with importlib_resources.as_file(ref) as path:
        suite = loader.discover(ref)
    runner = unittest.runner.TextTestRunner()  # verbosity=2)
    ret = not runner.run(suite).wasSuccessful()
    sys.exit(ret)
